"""
===========================
Interface Utility Functions
===========================

The functions defined here are used to support the interactive and command-line
interfaces for ``vivarium``.

"""
from datetime import datetime
import functools
from pathlib import Path
import sys

from loguru import logger

from vivarium.exceptions import VivariumError


def run_from_ipython() -> bool:
    """Taken from https://stackoverflow.com/questions/5376837/how-can-i-do-an-if-run-from-ipython-test-in-python"""
    try:
        __IPYTHON__
        return True
    except NameError:
        return False


def log_progress(sequence, every=None, size=None, name='Items'):
    """Taken from https://github.com/alexanderkuk/log-progress"""
    from ipywidgets import IntProgress, HTML, VBox
    from IPython.display import display

    is_iterator = False
    if size is None:
        try:
            size = len(sequence)
        except TypeError:
            is_iterator = True
    if size is not None:
        if every is None:
            if size <= 200:
                every = 1
            else:
                every = int(size / 200)     # every 0.5%
    else:
        assert every is not None, 'sequence is iterator, set every'

    if is_iterator:
        progress = IntProgress(min=0, max=1, value=1)
        progress.bar_style = 'info'
    else:
        progress = IntProgress(min=0, max=size, value=0)
    label = HTML()
    box = VBox(children=[label, progress])
    display(box)

    index = 0
    try:
        for index, record in enumerate(sequence, 1):
            if index == 1 or index % every == 0:
                if is_iterator:
                    label.value = '{name}: {index} / ?'.format(
                        name=name,
                        index=index
                    )
                else:
                    progress.value = index
                    label.value = u'{name}: {index} / {size}'.format(
                        name=name,
                        index=index,
                        size=size
                    )
            yield record
    except Exception as e:
        progress.bar_style = 'danger'
        raise
    else:
        progress.bar_style = 'success'
        progress.value = index
        label.value = "{name}: {index}".format(
            name=name,
            index=str(index or '?')
        )


class InteractiveError(VivariumError):
    """Error raised when the Interactive context is in an inconsistent state."""
    pass


def raise_if_not_setup(system_type):
    type_error_map = {
        'run': 'Simulation must be setup before it can be run',
        'value': 'Value pipeline configuration is not complete until the simulation is setup.',
        'event': 'Event configuration is not complete until the simulation is setup.',
        'component': 'Component configuration is not complete until the simulation is setup.',
        'population': 'No population exists until the simulation is setup.',
    }
    err_msg = type_error_map[system_type]

    def method_wrapper(context_method):

        @functools.wraps(context_method)
        def wrapped_method(*args, **kwargs):
            instance = args[0]
            if not instance._setup:
                raise InteractiveError(err_msg)
            return context_method(*args, **kwargs)

        return wrapped_method

    return method_wrapper


def get_output_root(results_directory, model_specification_file):
    launch_time = datetime.now().strftime("%Y_%m_%d_%H_%M_%S")
    model_name = Path(model_specification_file).stem
    output_root = Path(results_directory + f"/{model_name}/{launch_time}")
    return output_root


def add_logging_sink(sink, verbose, colorize=False, serialize=False):
    message_format = ('<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | '
                      '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> '
                      '- <level>{message}</level>')
    if verbose:
        logger.add(sink, colorize=colorize, level="DEBUG", format=message_format, serialize=serialize)
    else:
        logger.add(sink, colorize=colorize, level="ERROR", format=message_format, serialize=serialize)


def configure_logging_to_terminal(verbose):
    logger.remove(0)  # Clear default configuration
    add_logging_sink(sys.stdout, verbose, colorize=True)


def configure_logging_to_file(output_directory):
    master_log = output_directory / 'simulation.log'
    add_logging_sink(master_log, verbose=True)



import pytest
import numpy as np
import pandas as pd

from vivarium.framework.utilities import (from_yearly, to_yearly, rate_to_probability, probability_to_rate,
                                          collapse_nested_dict, import_by_path, handle_exceptions)


def test_from_yearly():
    one_month = pd.Timedelta(days=30.5)
    rate = 0.01
    new_rate = from_yearly(rate, one_month)
    assert round(new_rate, 5) == round(0.0008356164383561645, 5)


def test_to_yearly():
    one_month = pd.Timedelta(days=30.5)
    rate = 0.0008356164383561645
    new_rate = to_yearly(rate, one_month)
    assert round(new_rate, 5) == round(0.01, 5)


def test_rate_to_probability():
    rate = np.array([0.001])
    prob = rate_to_probability(rate)
    assert np.isclose(prob, 0.00099950016662497809)


def test_probability_to_rate():
    prob = np.array([0.00099950016662497809])
    rate = probability_to_rate(prob)
    assert np.isclose(rate, 0.001)


def test_rate_to_probability_symmetry():
    rate = np.array([0.0001])
    for _ in range(100):
        prob = rate_to_probability(rate)
        assert np.isclose(rate, probability_to_rate(prob))
        rate += (1-0.0001)/100.0


def test_rate_to_probability_vectorizability():
    rate = 0.001
    rate = np.array([rate]*100)
    prob = rate_to_probability(rate)
    assert round(prob[10], 5) == round(0.00099950016662497809, 5)
    assert round(np.sum(rate), 5) == round(np.sum(probability_to_rate(prob)), 5)


def test_collapse_nested_dict():
    source = {'a': {'b': {'c': 1, 'd': 2}}, 'e': 3}
    result = collapse_nested_dict(source)
    assert set(result) == {
            ('a.b.c', 1),
            ('a.b.d', 2),
            ('e', 3),
            }


def test_import_class_by_path():
    cls = import_by_path('collections.abc.Set')
    from collections.abc import Set
    assert cls is Set


def test_import_function_by_path():
    func = import_by_path('vivarium.framework.utilities.import_by_path')
    assert func is import_by_path


def test_bad_import_by_path():
    with pytest.raises(ImportError):
        import_by_path('junk.garbage.SillyClass')
    with pytest.raises(AttributeError):
        import_by_path('vivarium.framework.components.SillyClass')


class CustomException(Exception): pass
@pytest.mark.parametrize("test_input", [KeyboardInterrupt, RuntimeError, CustomException])
def test_handle_exceptions(test_input):
    def raise_me(ex):
        raise ex

    with pytest.raises(test_input):
        func = handle_exceptions(raise_me(test_input), None, False)
        func()



import pytest
import yaml

from vivarium.framework.configuration import (ConfigurationError, build_simulation_configuration,
                                              build_model_specification, validate_model_specification_file,
                                              _get_default_specification, DEFAULT_PLUGINS)


def test_get_default_specification_user_config(mocker, test_user_config):
    expand_user_mock = mocker.patch('vivarium.framework.configuration.Path.expanduser')
    expand_user_mock.return_value = test_user_config

    default_spec = _get_default_specification()

    assert expand_user_mock.called_once_with('~/vivarium.yaml')

    with test_user_config.open() as f:
        data = {'configuration': yaml.full_load(f)}

    data.update(DEFAULT_PLUGINS)
    data.update({'components': {}})

    assert default_spec.to_dict() == data


def test_get_default_specification_no_user_config(mocker, test_data_dir):
    user_config = test_data_dir / 'oh_no_nothing_here.yaml'

    expand_user_mock = mocker.patch('vivarium.framework.configuration.Path.expanduser')
    expand_user_mock.return_value = user_config

    default_spec = _get_default_specification()

    assert expand_user_mock.called_once_with('~/vivarium.yaml')
    data = {'components': {}, 'configuration': {}}
    data.update(DEFAULT_PLUGINS)

    assert default_spec.to_dict() == data


def test_validate_model_specification_failures(mocker, test_data_dir, test_spec):
    with pytest.raises(ConfigurationError):
        validate_model_specification_file('made_up_file.yaml')

    with pytest.raises(ConfigurationError):
        model_spec = test_data_dir / 'bad_model_specification.txt'
        validate_model_specification_file(model_spec)

    with test_spec.open() as f:
        spec_dict = yaml.full_load(f)
    spec_dict.update({'invalid_key': 'some_value'})
    load_mock = mocker.patch('vivarium.framework.configuration.yaml.full_load')
    load_mock.return_value = spec_dict
    with pytest.raises(ConfigurationError):
        validate_model_specification_file(test_spec)


def test_validate_model_specification(test_spec):
    validate_model_specification_file(test_spec)


def test_build_simulation_configuration(mocker, test_user_config):
    expand_user_mock = mocker.patch('vivarium.framework.configuration.Path.expanduser')
    expand_user_mock.return_value = test_user_config

    config = build_simulation_configuration()

    assert expand_user_mock.called_once_with('~/vivarium.yaml')

    with test_user_config.open() as f:
        data = yaml.full_load(f)

    assert config.to_dict() == data


def test_build_model_specification_failure(mocker, test_data_dir, test_spec):
    with pytest.raises(ConfigurationError):
        build_model_specification('made_up_file.yaml')

    with pytest.raises(ConfigurationError):
        model_spec = test_data_dir / 'bad_model_specification.txt'
        build_model_specification(model_spec)

    with test_spec.open() as f:
        spec_dict = yaml.full_load(f)
    spec_dict.update({'invalid_key': 'some_value'})
    load_mock = mocker.patch('vivarium.framework.configuration.yaml.full_load')
    load_mock.return_value = spec_dict
    with pytest.raises(ConfigurationError):
        build_model_specification(str(test_spec))


def test_build_model_specification(mocker, test_spec, test_user_config):
    expand_user_mock = mocker.patch('vivarium.framework.configuration.Path.expanduser')
    expand_user_mock.return_value = test_user_config
    loaded_model_spec = build_model_specification(test_spec)

    test_data = DEFAULT_PLUGINS.copy()

    with test_spec.open() as f:
        model_data = yaml.full_load(f)

    test_data.update(model_data)

    with test_user_config.open() as f:
        user_data = yaml.full_load(f)

    test_data['configuration'].update(user_data)

    assert loaded_model_spec.to_dict() == test_data



from pathlib import Path
import random
import pandas as pd

import pytest
from vivarium.testing_utilities import build_table, metadata
from vivarium.framework.artifact.manager import (_subset_rows, _subset_columns,
                                                 parse_artifact_path_config, ArtifactManager,
                                                 _config_filter, validate_filter_term)

@pytest.fixture()
def artifact_mock(mocker):
    mock = mocker.patch('vivarium.framework.artifact.manager.Artifact')

    def mock_load(key):
        if key == 'string_data.key':
            return 'string_data'
        elif key == 'df_data.key':
            return pd.DataFrame()
        else:
            return None

    mock.load.side_effect = mock_load

    return mock


def test_subset_rows_extra_filters():
    data = build_table(1, 1990, 2010)
    with pytest.raises(ValueError):
        _subset_rows(data, missing_thing=12)


def test_subset_rows():
    values = [lambda *args, **kwargs: random.choice(['red', 'blue']),
              lambda *args, **kwargs: random.choice([1, 2, 3])]
    data = build_table(values, 1990, 2010, columns=('age', 'year', 'sex', 'color', 'number'))

    filtered_data = _subset_rows(data, color='red', number=3)
    assert filtered_data.equals(data[(data.color == 'red') & (data.number == 3)])

    filtered_data = _subset_rows(data, color='red', number=[2, 3])
    assert filtered_data.equals(data[(data.color == 'red') & ((data.number == 2) | (data.number == 3))])


def test_subset_columns():
    values = [0, 'red', 100]
    data = build_table(values, 1990, 2010, columns=('age', 'year', 'sex', 'draw', 'color', 'value'))

    filtered_data = _subset_columns(data)
    assert filtered_data.equals(data[['age_start', 'age_end', 'year_start',
                                      'year_end', 'sex', 'color', 'value']])

    filtered_data = _subset_columns(data, color='red')
    assert filtered_data.equals(data[['age_start', 'age_end', 'year_start',
                                      'year_end', 'sex', 'value']])


def test_parse_artifact_path_config(base_config, test_data_dir):
    artifact_path = test_data_dir / 'artifact.hdf'
    base_config.update({'input_data': {'artifact_path': str(artifact_path)}}, **metadata(str(Path('/'))))

    assert parse_artifact_path_config(base_config) == str(artifact_path)


def test_parse_artifact_path_relative_no_source(base_config):
    artifact_path = './artifact.hdf'
    base_config.update({'input_data': {'artifact_path': str(artifact_path)}})

    with pytest.raises(ValueError):
        parse_artifact_path_config(base_config)


def test_parse_artifact_path_relative(base_config, test_data_dir):
    base_config.update({'input_data': {'artifact_path': '../../test_data/artifact.hdf'}},
                       **metadata(__file__))
    assert parse_artifact_path_config(base_config) == str(test_data_dir / 'artifact.hdf')


def test_parse_artifact_path_config_fail(base_config):
    artifact_path = Path(__file__).parent / 'not_an_artifact.hdf'
    base_config.update({'input_data': {'artifact_path': str(artifact_path)}}, **metadata(str(Path('/'))))

    with pytest.raises(FileNotFoundError):
        parse_artifact_path_config(base_config)


def test_parse_artifact_path_config_fail_relative(base_config):
    base_config.update({'input_data': {'artifact_path': './not_an_artifact.hdf'}}, **metadata(__file__))

    with pytest.raises(FileNotFoundError):
        parse_artifact_path_config(base_config)


def test_load_with_string_data(artifact_mock):
    am = ArtifactManager()
    am.artifact = artifact_mock
    am.config_filter_term = None
    assert am.load('string_data.key') == 'string_data'


def test_load_with_no_data(artifact_mock):
    am = ArtifactManager()
    am.artifact = artifact_mock
    assert am.load('no_data.key') is None


def test_load_with_df_data(artifact_mock):
    am = ArtifactManager()
    am.artifact = artifact_mock
    am.config_filter_term = None
    assert isinstance(am.load('df_data.key'), pd.DataFrame)


def test_config_filter():
    df = pd.DataFrame({'year': range(1990, 2000, 1), 'color': ['red', 'yellow']*5})
    filtered = +_config_filter(df, 'year in [1992, 1995]')

    assert set(filtered.year) == {1992, 1995}


def test_config_filter_on_nonexistent_column():
    df = pd.DataFrame({'year': range(1990, 2000, 1), 'color': ['red', 'yellow']*5})
    filtered = _config_filter(df, 'fake_col in [1992, 1995]')

    assert df.equals(filtered)


def test_validate_filter_term():
    multiple_filter_terms = 'draw in [0, 1] and year > 1990'

    with pytest.raises(NotImplementedError):
        validate_filter_term(multiple_filter_terms)



import pandas as pd

from vivarium.framework.state_machine import State, Machine, Transition
from vivarium.framework.utilities import rate_to_probability
from vivarium.framework.values import list_combiner, union_post_processor


class DiseaseTransition(Transition):

    def __init__(self, input_state, output_state, cause_key, measure, rate_name, **kwargs):
        super().__init__(input_state, output_state, probability_func=self._probability, **kwargs)
        self.cause_key = cause_key
        self.measure = measure
        self.rate_name = rate_name

    # noinspection PyAttributeOutsideInit
    def setup(self, builder):
        rate = builder.configuration[self.cause_key][self.measure]

        self.base_rate = lambda index: pd.Series(rate, index=index)
        self.transition_rate = builder.value.register_rate_producer(self.rate_name,
                                                                    source=self._risk_deleted_rate)
        self.joint_population_attributable_fraction = builder.value.register_value_producer(
            f'{self.rate_name}.population_attributable_fraction',
            source=lambda index: [pd.Series(0, index=index)],
            preferred_combiner=list_combiner,
            preferred_post_processor=union_post_processor)

    def _probability(self, index):
        effective_rate = self.transition_rate(index)
        return rate_to_probability(effective_rate)

    def _risk_deleted_rate(self, index):
        return self.base_rate(index) * (1 - self.joint_population_attributable_fraction(index))


class DiseaseState(State):

    def __init__(self, state_name, cause_key, with_excess_mortality=False):
        super().__init__(state_name)
        self.cause_key = cause_key
        self.with_excess_mortality = with_excess_mortality

    # noinspection PyAttributeOutsideInit
    def setup(self, builder):
        """Performs this component's simulation setup.

        Parameters
        ----------
        builder : `engine.Builder`
            Interface to several simulation tools.
        """
        super().setup(builder)
        if self.with_excess_mortality:
            self._excess_mortality_rate = builder.configuration[self.cause_key].excess_mortality_rate
        else:
            self._excess_mortality_rate = 0

        self.clock = builder.time.clock()

        self.excess_mortality_rate = builder.value.register_rate_producer(
            f'{self.state_id}.excess_mortality_rate',
            source=self.risk_deleted_excess_mortality_rate
        )
        self.excess_mortality_rate_paf = builder.value.register_value_producer(
            f'{self.state_id}.excess_mortality_rate.population_attributable_fraction',
            source=lambda index: [pd.Series(0, index=index)],
            preferred_combiner=list_combiner,
            preferred_post_processor=union_post_processor
        )

        builder.value.register_value_modifier('mortality_rate', self.add_in_excess_mortality)
        self.population_view = builder.population.get_view(
            [self._model], query=f"alive == 'alive' and {self._model} == '{self.state_id}'")

    def add_transition(self, output, measure, rate_name, **kwargs):
        t = DiseaseTransition(self, output, self.cause_key, measure, rate_name, **kwargs)
        self.transition_set.append(t)
        return t

    def risk_deleted_excess_mortality_rate(self, index):
        return pd.Series(self._excess_mortality_rate, index=index) * (1 - self.excess_mortality_rate_paf(index))

    def add_in_excess_mortality(self, index, mortality_rates):
        affected = self.population_view.get(index)
        mortality_rates.loc[affected.index] += self.excess_mortality_rate(affected.index)

        return mortality_rates


class DiseaseModel(Machine):

    def __init__(self, disease, initial_state, **kwargs):
        super().__init__(disease, **kwargs)
        self.initial_state = initial_state.state_id

    # noinspection PyAttributeOutsideInit
    def setup(self, builder):
        super().setup(builder)
        config = builder.configuration[self.state_column]
        # Reasonable approximation for short duration diseases.
        case_fatality_rate = config.excess_mortality_rate / (config.excess_mortality_rate + config.remission_rate)
        cause_specific_mortality_rate = config.incidence_rate * case_fatality_rate

        self.cause_specific_mortality_rate = builder.value.register_rate_producer(
            f'{self.state_column}.cause_specific_mortality_rate',
            source=lambda index: pd.Series(cause_specific_mortality_rate, index=index)
        )
        builder.value.register_value_modifier('mortality_rate', modifier=self.delete_cause_specific_mortality)
        builder.value.register_value_modifier('metrics', modifier=self.metrics)

        creates_columns = [self.state_column]
        builder.population.initializes_simulants(self.on_initialize_simulants,
                                                 creates_columns=creates_columns)
        self.population_view = builder.population.get_view(['age', 'sex', self.state_column])

        builder.event.register_listener('time_step', self.on_time_step)

    def on_initialize_simulants(self, pop_data):
        condition_column = pd.Series(self.initial_state, index=pop_data.index, name=self.state_column)
        self.population_view.update(condition_column)

    def on_time_step(self, event):
        self.transition(event.index, event.time)

    def delete_cause_specific_mortality(self, index, rates):
        return rates - self.cause_specific_mortality_rate(index)

    def metrics(self, index, metrics):
        pop = self.population_view.get(index, query="alive == 'alive'")
        metrics[self.state_column + '_prevalent_cases'] = len(pop[pop[self.state_column] != self.initial_state])
        return metrics


class SISDiseaseModel:

    configuration_defaults = {
        'disease': {
            'incidence_rate': 0.005,
            'remission_rate': 0.05,
            'excess_mortality_rate': 0.01,
        }
    }

    def __init__(self, disease_name):
        self._name = disease_name
        self.configuration_defaults = {disease_name: SISDiseaseModel.configuration_defaults['disease']}

        susceptible_state = DiseaseState(f'susceptible_to_{self._name}', self._name)
        infected_state = DiseaseState(f'infected_with_{self._name}', self._name, with_excess_mortality=True)

        susceptible_state.allow_self_transitions()
        susceptible_state.add_transition(infected_state, measure='incidence_rate',
                                         rate_name=f'{infected_state.state_id}.incidence_rate')
        infected_state.allow_self_transitions()
        infected_state.add_transition(susceptible_state, measure='remission_rate',
                                      rate_name=f'{infected_state.state_id}.remission_rate')

        model = DiseaseModel(self._name,
                             initial_state=susceptible_state,
                             states=[susceptible_state, infected_state])
        self._sub_components = [model]

    @property
    def name(self):
        return self._name

    @property
    def sub_components(self):
        return self._sub_components



import pandas as pd
import numpy as np
import pytest

from vivarium.framework.event import Event, EventManager


@pytest.fixture
def event_init():
    return {
        'orig': {
            'index': pd.Index(range(10)),
            'user_data': {'key': 'value'},
            'time': pd.Timestamp('1/1/2000'),
            'step_size': 12,
        },
        'new_val': {
            'index': pd.Index(range(3)),
            'user_data': {'new_key': 'new_value'},
            'time': pd.Timestamp.now(),
            'step_size': 30,
        }
    }


def test_proper_access(event_init):
    # Event attributes are meant to be read-only
    event_data = event_init['orig']
    e1 = Event(event_data['index'],
               event_data['user_data'],
               event_data['time'],
               event_data['step_size'])

    assert (event_data['index'] == e1.index).all()
    assert event_data['user_data'] == e1.user_data
    assert event_data['time'] == e1.time
    assert event_data['step_size'] == e1.step_size

    attribute_data = event_init['new_val']
    for key, value in attribute_data.items():
        with pytest.raises(AttributeError) as _:
            setattr(e1, key, value)


def test_split_event(event_init):
    event_data = event_init['orig']
    e1 = Event(event_data['index'],
               event_data['user_data'],
               event_data['time'],
               event_data['step_size'])

    new_idx = event_init['new_val']['index']
    e2 = e1.split(new_idx)

    assert e1.index is event_data['index']
    assert e2.index is new_idx


def test_emission(event_init):
    signal = [False]

    def listener(*_, **__):
        signal[0] = True

    manager = EventManager()
    manager.clock = lambda: pd.Timestamp(1990, 1, 1)
    manager.step_size = lambda: pd.Timedelta(30, unit='D')
    manager.add_constraint = lambda f, **kwargs: f
    emitter = manager.get_emitter('test_event')
    manager.register_listener('test_event', listener)
    emitter(event_init['orig']['index'])

    assert signal[0]

    signal[0] = False

    emitter = manager.get_emitter('test_unheard_event')
    emitter(event_init['new_val']['index'])
    assert not signal[0]


def test_listener_priority(event_init):
    signal = [False, False, False]

    def listener1(*_, **__):
        signal[0] = True
        assert not signal[1]
        assert not signal[2]

    def listener2(*_, **__):
        signal[1] = True
        assert signal[0]
        assert not signal[2]

    def listener3(*_, **__):
        signal[2] = True
        assert signal[0]
        assert signal[1]

    manager = EventManager()
    manager.clock = lambda: pd.Timestamp(1990, 1, 1)
    manager.step_size = lambda: pd.Timedelta(30, 'D')
    manager.add_constraint = lambda f, **kwargs: f
    emitter = manager.get_emitter('test_event')
    manager.register_listener('test_event', listener1, priority=0)
    manager.register_listener('test_event', listener2)
    manager.register_listener('test_event', listener3, priority=9)

    emitter(event_init['orig']['index'])
    assert np.all(signal)


def test_contains():
    event = 'test_event'

    manager = EventManager()
    manager.add_constraint = lambda f, **kwargs: f
    assert event not in manager
    manager.get_emitter(event)
    assert event in manager



from typing import Dict

import pandas as pd

from vivarium.framework.engine import Builder


class Observer:

    configuration_defaults = {
        'mortality': {
            'life_expectancy': 80,
        }
    }

    def __init__(self):
        self.name = 'observer'

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder):
        self.life_expectancy = builder.configuration.mortality.life_expectancy
        self.population_view = builder.population.get_view(['age', 'alive'])

        builder.value.register_value_modifier('metrics', self.metrics)

    def metrics(self, index: pd.Index, metrics: Dict):

        pop = self.population_view.get(index)
        metrics['total_population_alive'] = len(pop[pop.alive == 'alive'])
        metrics['total_population_dead'] = len(pop[pop.alive == 'dead'])

        metrics['years_of_life_lost'] = (self.life_expectancy - pop.age[pop.alive == 'dead']).sum()

        return metrics



"""
==================================
The Component Configuration Parser
==================================

The :class:`ComponentConfigurationParser` is responsible for taking a list or
hierarchical :class:`ConfigTree <vivarium.config_tree.ConfigTree>` of components
derived from a model specification yaml file and turning it into a list of
instantiated component objects. When a model specification yaml file is loaded,
the components come in as strings. In order for the simulation to be able to
use these components, they have to be converted into the actual objects they
represent. This occurs via the :meth:`get_components
<ComponentConfigurationParser.get_components>` method of the parser, which is
used anytime a simulation is initialized from a model specification file.

There are three steps to this process.

1. Parsing the model specification's components
2. Validating the arguments and prepping each component
3. Importing and instantiating the actual components

"""
from typing import Tuple, List, Dict, Union, Any

from vivarium.config_tree import ConfigTree
from vivarium.framework.utilities import import_by_path
from .manager import ComponentConfigError


class ParsingError(ComponentConfigError):
    """Error raised when component configurations are not specified correctly."""
    pass


class ComponentConfigurationParser:
    """Parses component configuration from model specification and initializes
    components.

    To define your own set of parsing rules, you should write a parser class
    that inherits from this class and overrides ``parse_component_config``.
    You can then define a set of parsing rules that turn component configuration
    information into a list of strings where each string is the full python
    import path to the class followed by a set of parentheses containing
    initialization arguments.

    For example, say your component configuration specifies ``FancyClass`` from
    the ``important_module`` of your ``made_up_package`` and that ``FancyClass``
    has two initialization parameters, ``'important_thing1'`` and
    ``'important_thing2'``. Your implementation of ``parse_component_config``
    needs to generate the string ``'made_up_package.important_module.FancyClass
    ("important_thing1", "important_thing2")'`` and include it in the list of
    components to generate.

    All classes that are initialized from the ``yaml`` configuration must
    either take no arguments or take arguments specified as strings.
    """

    def get_components(self, component_config: Union[ConfigTree, List]) -> List:
        """Extracts component specifications from configuration information and
        returns initialized components.

        This method encapsulates the three steps described above of parsing,
        validating/prepping, and importing/instantiating.

        The first step of parsing is only done for component configurations that
        come in as a :class:`ConfigTree <vivarium.config_tree.ConfigTree>`.
        Configurations that are provided in the form of a list are already
        assumed to be in the correct form.

        Parameters
        ----------
        component_config
            A hierarchical component specification blob. This configuration
            information needs to be parsable into a full import path and a set
            of initialization arguments by the ``parse_component_config``
            method.

        Returns
        -------
        List
            A list of initialized components.

        """
        if isinstance(component_config, ConfigTree):
            component_list = self.parse_component_config(component_config.to_dict())
        else:  # Components were specified in a list rather than a tree.
            component_list = component_config
        component_list = prep_components(component_list)
        return import_and_instantiate_components(component_list)

    def parse_component_config(self, component_config: Dict[str, Union[Dict, List]]) -> List[str]:
        """Parses a hierarchical component specification into a list of
        standardized component definitions.

        This default parser expects component configurations as a list of dicts.
        Each dict at the top level corresponds to a different package and has
        a single key. This key may be just the name of the package or a Python
        style import path to the module in which components live. The values of
        the top level dicts are a list of dicts or strings. If dicts, the keys
        are another step along the import path. If strings, the strings are
        representations of calls to the class constructor of components to be
        generated. This pattern may be arbitrarily nested.

        Parameters
        ----------
        component_config
            A hierarchical component specification blob.

        Returns
        -------
            A list of standardized component definitions. Component definition
            strings are specified as
            ``'absolute.import.path.ClassName("argument1", "argument2", ...)'``.

        """
        return parse_component_config_to_list(component_config)


def parse_component_config_to_list(component_config: Dict[str, Union[Dict, List]]) -> List[str]:
    """Helper function for parsing hierarchical component configuration into a
    flat list.

    This function recursively walks the component configuration dictionary,
    treating it like a prefix tree and building the import path prefix. When it
    hits a list, it prepends the built prefix onto each item in the list. If the
    dictionary contains multiple lists, the prefix-prepended lists are concated
    together. For example, a component configuration dictionary like the
    following:

    .. code-block:: python

        component_config = {
            'vivarium_examples': {
                'disease_model': {
                    'population': ['BasePopulation()', 'Mortality()']
                    'disease': ['SIS_DiseaseModel("diarrhea")]
                }
            }
        }

    would be parsed by this function into the following list:

    .. code-block:: python

        ['vivarium_examples.disease_model.population.BasePopulation()',
         'vivarium_examples.disease_model.population.Mortality()',
         'vivarium_examples.disease_model.disease.SIS_DiseaseModel("diarrhea")']


    Parameters
    ----------
    component_config
        A hierarchical component specification blob.

    Returns
    -------
        A flat list of strings, each string representing the full python import
        path to the component, the component name, and any arguments.
    """
    if not component_config:
        return []

    def _process_level(level, prefix):
        if not level:
            raise ParsingError(f'Check your configuration. Component {prefix} should not be left empty with the header')

        if isinstance(level, list):
            return ['.'.join(prefix + [child]) for child in level]

        component_list = []
        for name, child in level.items():
            component_list.extend(_process_level(child, prefix + [name]))

        return component_list

    return _process_level(component_config, [])


def prep_components(component_list: Union[List[str], Tuple[str]]) -> List[Tuple[str, Tuple]]:
    """Transform component description strings into tuples of component paths
    and required arguments.

    Parameters
    ----------
    component_list
        The component descriptions to transform.

    Returns
    -------
    List[Tuple[str, Tuple]]
        List of component/argument tuples.
    """
    components = []
    for c in component_list:
        path, args_plus = c.split('(')
        cleaned_args = clean_args(args_plus[:-1].split(','), path)
        components.append((path, cleaned_args))
    return components


def clean_args(args: List, path: str) -> Tuple:
    """Transform component arguments into a tuple, validating that each argument
    is a string.

    Parameters
    ----------
    args
        List of arguments to the component specified at ``path``.
    path
        Path representing the component for which arguments are being cleaned.

    Returns
    -------
    Tuple
        A tuple of arguments, each of which is guaranteed to be a string.
    """
    out = []
    for a in args:
        a = a.strip()
        if not a:
            continue

        not_a_valid_string = len(a) < 3 or not ((a[0] == a[-1] == "'") or (a[0] == a[-1] == '"'))
        if not_a_valid_string:
            raise ParsingError(f"Invalid component argument {a} for component {path}")

        out.append(a[1:-1])
    return tuple(out)


def import_and_instantiate_components(component_list: List[Tuple[str, Tuple[str]]]) -> List:
    """Transform the list of tuples representing components into the actual
    instantiated component objects.

    Parameters
    ----------
    component_list
        A list of tuples representing components, where the first element of
        each tuple is a string with the full import path to the component and
        the component name and the second element is a tuple of the arguments to
        the component.

    Returns
    -------
    List
        A list of instantiated component objects.

    """
    return [import_by_path(component)(*args) for component, args in component_list]



"""
============================
The Plugin Management System
============================

.. todo::
   This part will come in with the full description of the plugin system
   in the next PR. -J.C. 05/07/19

"""
from vivarium.config_tree import ConfigTree
from vivarium.exceptions import VivariumError

from .utilities import import_by_path

_MANAGERS = {
    'lookup': {
        'controller': 'vivarium.framework.lookup.LookupTableManager',
        'builder_interface': 'vivarium.framework.lookup.LookupTableInterface',
    },
    'randomness': {
        'controller': 'vivarium.framework.randomness.RandomnessManager',
        'builder_interface': 'vivarium.framework.randomness.RandomnessInterface',
    },
    'value': {
        'controller': 'vivarium.framework.values.ValuesManager',
        'builder_interface': 'vivarium.framework.values.ValuesInterface',
    },
    'event': {
        'controller': 'vivarium.framework.event.EventManager',
        'builder_interface': 'vivarium.framework.event.EventInterface',
    },
    'population': {
        'controller': 'vivarium.framework.population.PopulationManager',
        'builder_interface': 'vivarium.framework.population.PopulationInterface',
    },
    'resource': {
        'controller': 'vivarium.framework.resource.ResourceManager',
        'builder_interface': 'vivarium.framework.resource.ResourceInterface',
    },
    'results': {
        'controller': 'vivarium.framework.results.ResultsManager',
        'builder_interface': 'vivarium.framework.results.ResultsInterface',
    },
}

DEFAULT_PLUGINS = {
    'plugins': {
        'required': {
            'component_manager': {
                'controller': 'vivarium.framework.components.ComponentManager',
                'builder_interface': 'vivarium.framework.components.ComponentInterface',
            },
            'clock': {
                'controller': 'vivarium.framework.time.DateTimeClock',
                'builder_interface': 'vivarium.framework.time.TimeInterface',
            },
            'component_configuration_parser': {
                'controller': 'vivarium.framework.components.ComponentConfigurationParser',
                'builder_interface': None,
            },
            'lifecycle': {
                'controller': 'vivarium.framework.lifecycle.LifeCycleManager',
                'builder_interface': 'vivarium.framework.lifecycle.LifeCycleInterface',
            },
            'data': {
                'controller': 'vivarium.framework.artifact.ArtifactManager',
                'builder_interface': 'vivarium.framework.artifact.ArtifactInterface'
            }
        },
        'optional': {}
    }
}


class PluginConfigurationError(VivariumError):
    """Error raised when plugin configuration is incorrectly specified."""
    pass


class PluginManager:

    def __init__(self, plugin_configuration=None):
        self._plugin_configuration = ConfigTree(DEFAULT_PLUGINS['plugins'], layers=['base', 'override'])
        self._plugin_configuration.update(plugin_configuration, source='initialization_args')
        self._plugins = {}

    def get_plugin(self, name):
        if name not in self._plugins:
            self._plugins[name] = self._get(name)
        return self._plugins[name]['controller']

    def get_plugin_interface(self, name):
        if name not in self._plugins:
            self._plugins[name] = self._get(name)
        return self._plugins[name]['builder_interface']

    def get_core_controllers(self):
        core_components = [name for name in self._plugin_configuration['required'].keys()] + list(_MANAGERS.keys())
        return {name: self.get_plugin(name) for name in core_components}

    def get_core_interfaces(self):
        core_components = [name for name in self._plugin_configuration['required'].keys()] + list(_MANAGERS.keys())
        return {name: self.get_plugin_interface(name) for name in core_components}

    def get_optional_controllers(self):
        return {name: self.get_plugin(name) for name in self._plugin_configuration['optional'].keys()}

    def get_optional_interfaces(self):
        return {name: self.get_plugin_interface(name) for name in self._plugin_configuration['optional'].keys()}

    def _get(self, name):
        if name not in self._plugins:
            self._plugins[name] = self._build_plugin(name)
        return self._plugins[name]

    def _build_plugin(self, name):
        plugin = self._lookup(name)

        try:
            controller = import_by_path(plugin['controller'])()
        except ValueError:
            raise PluginConfigurationError(f'Invalid plugin specification {plugin["controller"]}')

        if plugin['builder_interface'] is not None:
            try:
                interface = import_by_path(plugin['builder_interface'])(controller)
            except ValueError:
                raise PluginConfigurationError(f'Invalid plugin specification {plugin["builder_interface"]}')
        else:
            interface = None

        return {'controller': controller, 'builder_interface': interface}

    def _lookup(self, name):
        if name in self._plugin_configuration['required']:
            return self._plugin_configuration['required'][name]
        elif name in self._plugin_configuration['optional']:
            return self._plugin_configuration['optional'][name]
        elif name in _MANAGERS:
            return _MANAGERS[name]
        else:
            raise PluginConfigurationError(f'Plugin {name} not found.')

    def __repr__(self):
        return "PluginManager()"



import pandas as pd

from vivarium.framework.engine import Builder
from vivarium.framework.population import SimulantData
from vivarium.framework.event import Event


class BasePopulation:
    """Generates a base population with a uniform distribution of age and sex.

    Attributes
    ----------
    configuration_defaults :
        A set of default configuration values for this component. These can be
        overwritten in the simulation model specification or by providing
        override values when constructing an interactive simulation.
    """

    configuration_defaults = {
        'population': {
            # The range of ages to be generated in the initial population
            'age_start': 0,
            'age_end': 100,
            # Note: There is also a 'population_size' key.
        },
    }

    def __init__(self):
        self.name = 'base_population'

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder):
        """Performs this component's simulation setup.

        The ``setup`` method is automatically called by the simulation
        framework. The framework passes in a ``builder`` object which
        provides access to a variety of framework subsystems and metadata.

        Parameters
        ----------
        builder :
            Access to simulation tools and subsystems.
        """
        self.config = builder.configuration

        self.with_common_random_numbers = bool(self.config.randomness.key_columns)
        self.register = builder.randomness.register_simulants
        if (self.with_common_random_numbers
                and not ['entrance_time', 'age'] == self.config.randomness.key_columns):
            raise ValueError("If running with CRN, you must specify ['entrance_time', 'age'] as"
                             "the randomness key columns.")

        self.age_randomness = builder.randomness.get_stream('age_initialization',
                                                            for_initialization=self.with_common_random_numbers)
        self.sex_randomness = builder.randomness.get_stream('sex_initialization')

        columns_created = ['age', 'sex', 'alive', 'entrance_time']
        builder.population.initializes_simulants(self.on_initialize_simulants,
                                                 creates_columns=columns_created)

        self.population_view = builder.population.get_view(columns_created)

        builder.event.register_listener('time_step', self.age_simulants)

    def on_initialize_simulants(self, pop_data: SimulantData):
        """Called by the simulation whenever new simulants are added.

        This component is responsible for creating and filling four columns
        in the population state table:

        'age' :
            The age of the simulant in fractional years.
        'sex' :
            The sex of the simulant. One of {'Male', 'Female'}
        'alive' :
            Whether or not the simulant is alive. One of {'alive', 'dead'}
        'entrance_time' :
            The time that the simulant entered the simulation. The 'birthday'
            for simulants that enter as newborns. A `pandas.Timestamp`.

        Parameters
        ----------
        pop_data :
            A record containing the index of the new simulants, the
            start of the time step the simulants are added on, the width
            of the time step, and the age boundaries for the simulants to
            generate.

        """

        age_start = self.config.population.age_start
        age_end = self.config.population.age_end
        if age_start == age_end:
            age_window = pop_data.creation_window / pd.Timedelta(days=365)
        else:
            age_window = age_end - age_start

        age_draw = self.age_randomness.get_draw(pop_data.index)
        age = age_start + age_draw * age_window

        if self.with_common_random_numbers:
            population = pd.DataFrame({'entrance_time': pop_data.creation_time,
                                       'age': age.values}, index=pop_data.index)
            self.register(population)
            population['sex'] = self.sex_randomness.choice(pop_data.index, ['Male', 'Female'])
            population['alive'] = 'alive'
        else:
            population = pd.DataFrame(
                {'age': age.values,
                 'sex': self.sex_randomness.choice(pop_data.index, ['Male', 'Female']),
                 'alive': pd.Series('alive', index=pop_data.index),
                 'entrance_time': pop_data.creation_time},
                index=pop_data.index)

        self.population_view.update(population)

    def age_simulants(self, event: Event):
        """Updates simulant age on every time step.

        Parameters
        ----------
        event :
            An event object emitted by the simulation containing an index
            representing the simulants affected by the event and timing
            information.
        """
        population = self.population_view.get(event.index, query="alive == 'alive'")
        population['age'] += event.step_size / pd.Timedelta(days=365)
        self.population_view.update(population)



import numpy
numpy.seterr(all='raise')

from vivarium.config_tree import ConfigTree
from vivarium.interface import InteractiveContext
from vivarium.framework.artifact import Artifact
from vivarium.framework.configuration import build_model_specification
from vivarium.__about__ import (__author__, __copyright__, __email__, __license__,
                                __summary__, __title__, __uri__, __version__, )



from .population import BasePopulation
from .mortality import Mortality
from .observer import Observer
from .disease import DiseaseModel, DiseaseState, DiseaseTransition, SISDiseaseModel
from .risk import Risk, RiskEffect
from .intervention import TreatmentIntervention


def get_model_specification_path():
    from pathlib import Path
    p = Path(__file__).parent / 'disease_model.yaml'
    return str(p)


def get_disease_model_simulation():
    from vivarium import InteractiveContext
    p = get_model_specification_path()
    return InteractiveContext(p)



"""
==============================
Random Numbers in ``vivarium``
==============================

This module contains classes and functions supporting common random numbers.

Vivarium has some peculiar needs around randomness. We need to be totally
consistent between branches in a comparison. For example, if a simulant gets
hit by a truck in the base case it must be hit by that same truck in the
counter-factual at exactly the same moment unless the counter-factual
explicitly deals with traffic accidents. That means that the system can't rely
on standard global randomness sources because small changes to the number of
bits consumed or the order in which randomness consuming operations occur will
cause the system to diverge.

The current approach is to generate hash-based
seeds where the key is the simulation time, the simulant's id, the draw number
and a unique id for the decision point which needs the randomness.
These seeds are then used to generate `numpy.random.RandomState` objects that
can be used to create pseudo-random numbers in a repeatable manner.

Attributes
----------
RESIDUAL_CHOICE : object
    A probability placeholder to be used in an un-normalized array of weights
    to absorb leftover weight so that the array sums to unity.
    For example::

        [0.2, 0.2, RESIDUAL_CHOICE] => [0.2, 0.2, 0.6]

    Note
    ----
    Currently this object is only used in the `choice` function of this
    module.


For mor information, see the Common Random Numbers
:ref:`concept note <crn_concept>`.

"""
from typing import Union, List, Tuple, Callable, Any
import hashlib
import datetime

import numpy as np
import pandas as pd

from vivarium.exceptions import VivariumError
from .utilities import rate_to_probability

Array = Union[List, Tuple, np.ndarray, pd.Series]
Index = Union[pd.Index, pd.MultiIndex]


class RandomnessError(VivariumError):
    """Raised for inconsistencies in random number and choice generation."""
    pass


RESIDUAL_CHOICE = object()


class IndexMap:
    """A key-index mapping with a vectorized hash and vectorized lookups."""

    TEN_DIGIT_MODULUS = 10_000_000_000

    def __init__(self, map_size=1_000_000):
        self._map = pd.Series(dtype=float)
        self.map_size = map_size

    def update(self, new_keys: Index):
        """Adds the new keys to the mapping.

        Parameters
        ----------
        new_keys
            The new index to hash.

        """
        if not self._map.index.intersection(new_keys).empty:
            raise KeyError("Non-unique keys in index.")

        mapping_update = self.hash_(new_keys)
        if self._map.empty:
            self._map = mapping_update.drop_duplicates()
        else:
            self._map = self._map.append(mapping_update).drop_duplicates()

        collisions = mapping_update.index.difference(self._map.index)
        salt = 1
        while not collisions.empty:
            mapping_update = self.hash_(collisions, salt)
            self._map = self._map.append(mapping_update).drop_duplicates()
            collisions = mapping_update.index.difference(self._map.index)
            salt += 1

    def hash_(self, keys: Index, salt: int = 0) -> pd.Series:
        """Hashes the index into an integer index in the range [0, self.stride]

        Parameters
        ----------
        keys
            The new index to hash.
        salt
            An integer used to perturb the hash in a deterministic way. Useful
            in dealing with collisions.

        Returns
        -------
        pandas.Series
            A pandas series indexed by the given keys and whose values take on
            integers in the range [0, self.stride].  Duplicates may appear and
            should be dealt with by the calling code.

        """
        key_frame = keys.to_frame()
        new_map = pd.Series(0, index=keys)
        salt = self.convert_to_ten_digit_int(pd.Series(salt, index=keys))

        for i, column_name in enumerate(key_frame.columns):
            column = self.convert_to_ten_digit_int(key_frame[column_name])

            primes = [2, 3, 5, 7, 11, 13, 17, 19, 23, 27]
            out = pd.Series(1, index=column.index)
            for idx, p in enumerate(primes):
                # numpy will almost always overflow here, but it is equivalent
                # to modding out by 2**64.  Since it's much much larger than
                # our map size the amount of additional periodicity this
                # introduces is pretty trivial.
                out *= np.power(p, self.digit(column, idx))
            new_map += out + salt

        return new_map % self.map_size

    def convert_to_ten_digit_int(self, column: pd.Series) -> pd.Series:
        """Converts a column of datetimes, integers, or floats into a column
        of 10 digit integers.

        Parameters
        ----------
        column
            A series of datetimes, integers, or floats.

        Returns
        -------
        pandas.Series
            A series of ten digit integers based on the input data.

        Raises
        ------
        RandomnessError
            If the column contains data that is neither a datetime-like nor
            numeric.

        """
        if isinstance(column.iloc[0], datetime.datetime):
            column = self.clip_to_seconds(column.view(np.int64))
        elif np.issubdtype(column.iloc[0], np.integer):
            if not len(column >= 0) == len(column):
                raise RandomnessError("Values in integer columns must be greater than or equal to zero.")
            column = self.spread(column)
        elif np.issubdtype(column.iloc[0], np.floating):
            column = self.shift(column)
        else:
            raise RandomnessError(f"Unhashable column type {type(column.iloc[0])}. "
                                  "IndexMap accepts datetime like columns and numeric columns.")
        return column

    @staticmethod
    def digit(m: Union[int, pd.Series], n: int) -> Union[int, pd.Series]:
        """Returns the nth digit of each number in m."""
        return (m // (10 ** n)) % 10

    @staticmethod
    def clip_to_seconds(m: Union[int, pd.Series]) -> Union[int, pd.Series]:
        """Clips UTC datetime in nanoseconds to seconds."""
        return m // pd.Timedelta(1, unit='s').value

    def spread(self, m: Union[int, pd.Series]) -> Union[int, pd.Series]:
        """Spreads out integer values to give smaller values more weight."""
        return (m * 111_111) % self.TEN_DIGIT_MODULUS

    def shift(self, m: Union[float, pd.Series]) -> Union[int, pd.Series]:
        """Shifts floats so that the first 10 decimal digits are significant."""
        out = m % 1 * self.TEN_DIGIT_MODULUS // 1
        if isinstance(out, pd.Series):
            return out.astype('int64')
        return int(out)

    def __getitem__(self, index: Index) -> pd.Series:
        if isinstance(index, (pd.Index, pd.MultiIndex)):
            return self._map[index]
        else:
            raise IndexError(index)

    def __len__(self) -> int:
        return len(self._map)

    def __repr__(self) -> str:
        return 'IndexMap({})'.format("\n         ".join(repr(self._map).split("\n")))


def random(key: str, index: Index, index_map: IndexMap = None) -> pd.Series:
    """Produces an indexed set of uniformly distributed random numbers.

    The index passed in typically corresponds to a subset of rows in a
    `pandas.DataFrame` for which a probabilistic draw needs to be made.

    Parameters
    ----------
    key :
        A string used to create a seed for the random number generation.
    index :
        The index used for the returned series.
    index_map :
        A mapping between the provided index (which may contain ints, floats,
        datetimes or any arbitrary combination of them) and an integer index
        into the random number array.

    Returns
    -------
    pandas.Series
        A series of random numbers indexed by the provided index.

    """
    if len(index) > 0:
        random_state = np.random.RandomState(seed=get_hash(key))

        # Generate a random number for every simulant.
        #
        # NOTE: We generate a full set of random numbers for the population
        # even when we may only need a few.  This ensures consistency in outcomes
        # across simulations.
        # See Also:
        # 1. https://en.wikipedia.org/wiki/Variance_reduction
        # 2. Untangling Uncertainty with Common Random Numbers: A Simulation Study; A.Flaxman, et. al., Summersim 2017
        sample_size = index_map.map_size if index_map is not None else index.max() + 1
        try:
            draw_index = index_map[index]
        except (IndexError, TypeError):
            draw_index = index
        raw_draws = random_state.random_sample(sample_size)
        return pd.Series(raw_draws[draw_index], index=index)

    return pd.Series(index=index)  # Structured null value


def get_hash(key: str) -> int:
    """Gets a hash of the provided key.

    Parameters
    ----------
    key :
        A string used to create a seed for the random number generator.

    Returns
    -------
    int
        A hash of the provided key.

    """
    # 4294967295 == 2**32 - 1 which is the maximum allowable seed for a `numpy.random.RandomState`.
    return int(hashlib.sha1(key.encode('utf8')).hexdigest(), 16) % 4294967295


def choice(key: str, index: Index, choices: Array, p: Array = None, index_map: IndexMap = None) -> pd.Series:
    """Decides between a weighted or unweighted set of choices.

    Given a a set of choices with or without corresponding weights,
    returns an indexed set of decisions from those choices. This is
    simply a vectorized way to make decisions with some book-keeping.

    Parameters
    ----------
    key
        A string used to create a seed for the random number generation.
    index
        An index whose length is the number of random draws made
        and which indexes the returned `pandas.Series`.
    choices
        A set of options to choose from.
    p
        The relative weights of the choices.  Can be either a 1-d array of
        the same length as `choices` or a 2-d array with `len(index)` rows
        and `len(choices)` columns.  In the 1-d case, the same set of weights
        are used to decide among the choices for every item in the `index`.
        In the 2-d case, each row in `p` contains a separate set of weights
        for every item in the `index`.
    index_map
        A mapping between the provided index (which may contain ints, floats,
        datetimes or any arbitrary combination of them) and an integer index
        into the random number array.

    Returns
    -------
    pandas.Series
        An indexed set of decisions from among the available `choices`.

    Raises
    ------
    RandomnessError
        If any row in `p` contains `RESIDUAL_CHOICE` and the remaining
        weights in the row are not normalized or any row of `p` contains
        more than one reference to `RESIDUAL_CHOICE`.

    """
    # Convert p to normalized probabilities broadcasted over index.
    p = _set_residual_probability(_normalize_shape(p, index)) if p is not None else np.ones((len(index), len(choices)))
    p = p/p.sum(axis=1, keepdims=True)

    draw = random(key, index, index_map)

    p_bins = np.cumsum(p, axis=1)
    # Use the random draw to make a choice for every row in index.
    choice_index = (draw.values[np.newaxis].T > p_bins).sum(axis=1)

    return pd.Series(np.array(choices)[choice_index], index=index)


def _normalize_shape(p: Array, index: Index) -> np.ndarray:
    p = np.array(p)
    # We got a 1-d array => same weights for every index.
    if len(p.shape) == 1:
        # Turn 1-d array into 2-d array with same weights in every row.
        p = np.array(np.broadcast_to(p, (len(index), p.shape[0])))
    return p


def _set_residual_probability(p: np.ndarray) -> np.ndarray:
    """Turns any use of `RESIDUAL_CHOICE` into a residual probability.

    Parameters
    ----------
    p
        Array where each row is a set of probability weights and potentially
        a `RESIDUAL_CHOICE` placeholder.

    Returns
    -------
    numpy.ndarray
        Array where each row is a set of normalized probability weights.

    """
    residual_mask = p == RESIDUAL_CHOICE
    if residual_mask.any():  # I.E. if we have any placeholders.
        if np.any(np.sum(residual_mask, axis=1) - 1):
            raise RandomnessError(
                'More than one residual choice supplied for a single set of weights. Weights: {}.'.format(p))

        p[residual_mask] = 0
        residual_p = 1 - np.sum(p, axis=1)  # Probabilities sum to 1.

        if np.any(residual_p < 0):  # We got un-normalized probability weights.
            raise RandomnessError(
                'Residual choice supplied with weights that summed to more than 1. Weights: {}.'.format(p))

        p[residual_mask] = residual_p
    return p


def filter_for_probability(key: str, population: Union[pd.DataFrame, pd.Series, Index],
                           probability: Array, index_map: IndexMap = None) -> Union[pd.DataFrame, pd.Series, Index]:
    """Decide an event outcome for each individual in a population from
    probabilities.

    Given a population or its index and an array of associated probabilities
    for some event to happen, we create and return the sub-population for whom
    the event occurred.

    Parameters
    ----------
    key :
        A string used to create a seed for the random number generation.
    population :
        A view on the simulants for which we are determining the
        outcome of an event.
    probability :
        A 1d list of probabilities of the event under consideration
        occurring which corresponds (i.e. `len(population) == len(probability)`)
        to the population array passed in.
    index_map :
        A mapping between the provided index (which may contain ints, floats,
        datetimes or any arbitrary combination of them) and an integer index
        into the random number array.

    Returns
    -------
    pandas.core.generic.PandasObject
        The sub-population of the simulants for whom the event occurred.
        The return type will be the same as type(population)

    """
    if population.empty:
        return population

    index = population if isinstance(population, pd.Index) else population.index
    draw = random(key, index, index_map)
    mask = np.array(draw < probability)
    return population[mask]


class RandomnessStream:
    """A stream for producing common random numbers.

    `RandomnessStream` objects provide an interface to Vivarium's
    common random number generation.  They provide a number of methods
    for doing common simulation tasks that require random numbers like
    making decisions among a number of choices.

    Attributes
    ----------
    key
        The name of the randomness stream.
    clock
        A way to get the current simulation time.
    seed
        An extra number used to seed the random number generation.

    Notes
    -----
    Should not be constructed by client code.

    Simulation components get `RandomnessStream` objects by requesting
    them from the builder provided to them during the setup phase.
    I.E.::

        class VivariumComponent:
            def setup(self, builder):
                self.randomness_stream = builder.randomness.get_stream('stream_name')

    """
    def __init__(self, key: str, clock: Callable, seed: Any,
                 index_map: IndexMap = None, manager: 'RandomnessManager' = None,
                 for_initialization: bool = False):
        self.key = key
        self.clock = clock
        self.seed = seed
        self.index_map = index_map
        self._manager = manager
        self._for_initialization = for_initialization

    def copy_with_additional_key(self, key: Any) -> 'RandomnessStream':
        """Creates a copy of this stream with a permutation of it's random seed.

        Parameters
        ----------
        key
            The additional key to describe the new stream with.

        Returns
        -------
        RandomnessStream
            A new RandomnessStream with a combined key.

        """
        copy_key = '_'.join([self.key, key])
        if self._for_initialization:
            raise RandomnessError('Initialization streams cannot be copied.')
        elif self._manager:
            return self._manager.get_randomness_stream(copy_key)
        else:
            return RandomnessStream(copy_key, self.clock, self.seed, self.index_map)

    @property
    def name(self):
        return f'randomness_stream_{self.key}'

    def _key(self, additional_key: Any = None) -> str:
        """Construct a hashable key from this object's state.

        Parameters
        ----------
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
        str
            A key to seed random number generation.

        """
        return '_'.join([self.key, str(self.clock()), str(additional_key), str(self.seed)])

    def get_draw(self, index: Index, additional_key: Any = None) -> pd.Series:
        """Get an indexed set of numbers uniformly drawn from the unit interval.

        Parameters
        ----------
        index
            An index whose length is the number of random draws made
            and which indexes the returned `pandas.Series`.
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
        pandas.Series
            A series of random numbers indexed by the provided `pandas.Index`.
        """
        if self._for_initialization:
            draw = random(self._key(additional_key), pd.Index(range(len(index))), self.index_map)
            draw.index = index
        else:
            draw = random(self._key(additional_key), index, self.index_map)

        return draw

    def filter_for_rate(self, population: Union[pd.DataFrame, pd.Series, Index],
                        rate: Array, additional_key: Any = None) -> Union[pd.DataFrame, pd.Series, Index]:
        """Decide an event outcome for each individual from rates.

        Given a population or its index and an array of associated rates for
        some event to happen, we create and return the sub-population for whom
        the event occurred.

        Parameters
        ----------
        population
            A view on the simulants for which we are determining the
            outcome of an event.
        rate
            A 1d list of rates of the event under consideration occurring which
            corresponds (i.e. `len(population) == len(probability))` to the
            population view passed in. The rates must be scaled to the
            simulation time-step size either manually or as a post-processing
            step in a rate pipeline.
        additional_key
            Any additional information used to create the seed.

        Returns
        -------
        pandas.core.generic.PandasObject
            The sub-population of the simulants for whom the event occurred.
            The return type will be the same as type(population)

        """
        return self.filter_for_probability(population, rate_to_probability(rate), additional_key)

    def filter_for_probability(self, population: Union[pd.DataFrame, pd.Series, Index],
                               probability: Array, additional_key: Any = None) -> Union[pd.DataFrame, pd.Series, Index]:
        """Decide an outcome for each individual from probabilities.

        Given a population or its index and an array of associated probabilities
        for some event to happen, we create and return the sub-population for
        whom the event occurred.

        Parameters
        ----------
        population
            A view on the simulants for which we are determining the
            outcome of an event.
        probability
            A 1d list of probabilities of the event under consideration
            occurring which corresponds (i.e.
            `len(population) == len(probability)` to the population view
            passed in.
        additional_key
            Any additional information used to create the seed.

        Returns
        -------
        pandas.core.generic.PandasObject
            The sub-population of the simulants for whom the event occurred.
            The return type will be the same as type(population)

        """
        return filter_for_probability(self._key(additional_key), population, probability, self.index_map)

    def choice(self, index: Index, choices: Array, p: Array = None, additional_key: Any = None) -> pd.Series:
        """Decides between a weighted or unweighted set of choices.

        Given a a set of choices with or without corresponding weights,
        returns an indexed set of decisions from those choices. This is
        simply a vectorized way to make decisions with some book-keeping.

        Parameters
        ----------
        index
            An index whose length is the number of random draws made
            and which indexes the returned `pandas.Series`.
        choices
            A set of options to choose from.
        p
            The relative weights of the choices.  Can be either a 1-d array of
            the same length as `choices` or a 2-d array with `len(index)` rows
            and `len(choices)` columns.  In the 1-d case, the same set of
            weights are used to decide among the choices for every item in
            the `index`. In the 2-d case, each row in `p` contains a separate
            set of weights for every item in the `index`.
        additional_key
            Any additional information used to seed random number generation.

        Returns
        -------
        pandas.Series
            An indexed set of decisions from among the available `choices`.

        Raises
        ------
        RandomnessError
            If any row in `p` contains `RESIDUAL_CHOICE` and the remaining
            weights in the row are not normalized or any row of `p contains
            more than one reference to `RESIDUAL_CHOICE`.

        """
        return choice(self._key(additional_key), index, choices, p, self.index_map)

    def __repr__(self) -> str:
        return "RandomnessStream(key={!r}, clock={!r}, seed={!r})".format(self.key, self.clock(), self.seed)


class RandomnessManager:
    """Access point for common random number generation."""

    configuration_defaults = {
        'randomness':
            {
                'map_size': 1_000_000,
                'key_columns': ['entrance_time'],
                'random_seed': 0,
                'additional_seed': None,
            }
    }

    def __init__(self):
        self._seed = None
        self._clock = None
        self._key_columns = None
        self._key_mapping = IndexMap()
        self._decision_points = dict()

    @property
    def name(self):
        return "randomness_manager"

    def setup(self, builder):
        self._seed = str(builder.configuration.randomness.random_seed)
        if builder.configuration.randomness.additional_seed is not None:
            self._seed += str(builder.configuration.randomness.additional_seed)
        self._clock = builder.time.clock()
        self._key_columns = builder.configuration.randomness.key_columns
        map_size = builder.configuration.randomness.map_size
        pop_size = builder.configuration.population.population_size
        self._key_mapping.map_size = max(map_size, 10*pop_size)

        self.resources = builder.resources
        self._add_constraint = builder.lifecycle.add_constraint
        self._add_constraint(self.get_seed, restrict_during=['initialization'])
        self._add_constraint(self.get_randomness_stream, allow_during=['setup'])
        self._add_constraint(self.register_simulants,
                             restrict_during=['initialization', 'setup', 'post_setup', 'simulation_end', 'report'])

    def get_randomness_stream(self, decision_point: str, for_initialization: bool = False) -> RandomnessStream:
        """Provides a new source of random numbers for the given decision point.

        Parameters
        ----------
        decision_point
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.
        for_initialization
            A flag indicating whether this stream is used to generate key
            initialization information that will be used to identify simulants
            in the Common Random Number framework. These streams cannot be
            copied and should only be used to generate the state table columns
            specified in ``builder.configuration.randomness.key_columns``.

        Raises
        ------
        RandomnessError
            If another location in the simulation has already created a randomness stream
            with the same identifier.

        """
        stream = self._get_randomness_stream(decision_point, for_initialization)
        if not for_initialization:  # We need the key columns to be created before this stream can be called.
            self.resources.add_resources('stream', [decision_point], stream,
                                         [f'column.{name}' for name in self._key_columns])
        self._add_constraint(stream.get_draw, restrict_during=['initialization', 'setup', 'post_setup'])
        self._add_constraint(stream.filter_for_probability, restrict_during=['initialization', 'setup', 'post_setup'])
        self._add_constraint(stream.filter_for_rate, restrict_during=['initialization', 'setup', 'post_setup'])
        self._add_constraint(stream.choice, restrict_during=['initialization', 'setup', 'post_setup'])

        return stream

    def _get_randomness_stream(self, decision_point: str, for_initialization: bool = False) -> RandomnessStream:
        if decision_point in self._decision_points:
            raise RandomnessError(f"Two separate places are attempting to create "
                                  f"the same randomness stream for {decision_point}")
        stream = RandomnessStream(key=decision_point, clock=self._clock, seed=self._seed,
                                  index_map=self._key_mapping, manager=self, for_initialization=for_initialization)
        self._decision_points[decision_point] = stream
        return stream

    def get_seed(self, decision_point: str) -> int:
        """Get a randomly generated seed for use with external randomness tools.

        Parameters
        ----------
        decision_point
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.

        Returns
        -------
        int
            A seed for a random number generation that is linked to Vivarium's
            common random number framework.

        """
        return get_hash('_'.join([decision_point, str(self._clock()), str(self._seed)]))

    def register_simulants(self, simulants: pd.DataFrame):
        """Adds new simulants to the randomness mapping.

        Parameters
        ----------
        simulants
            A table with state data representing the new simulants.  Each
            simulant should pass through this function exactly once.

        Raises
        ------
        RandomnessError
            If the provided table does not contain all key columns specified
            in the configuration.

        """
        if not all(k in simulants.columns for k in self._key_columns):
            raise RandomnessError("The simulants dataframe does not have all specified key_columns.")
        self._key_mapping.update(simulants.set_index(self._key_columns).index)

    def __str__(self):
        return "RandomnessManager()"

    def __repr__(self) -> str:
        return f"RandomnessManager(seed={self._seed}, key_columns={self._key_columns})"


class RandomnessInterface:

    def __init__(self, manager: RandomnessManager):
        self._manager = manager

    def get_stream(self, decision_point: str, for_initialization: bool = False) -> RandomnessStream:
        """Provides a new source of random numbers for the given decision point.

        ``vivarium`` provides a framework for Common Random Numbers which
        allows for variance reduction when modeling counter-factual scenarios.
        Users interested in causal analysis and comparisons between simulation
        scenarios should be careful to use randomness streams provided by the
        framework wherever randomness is employed.

        Parameters
        ----------
        decision_point
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.
        for_initialization
            A flag indicating whether this stream is used to generate key
            initialization information that will be used to identify simulants
            in the Common Random Number framework. These streams cannot be
            copied and should only be used to generate the state table columns
            specified in ``builder.configuration.randomness.key_columns``.

        Returns
        -------
        RandomnessStream
            An entry point into the Common Random Number generation framework.
            The stream provides vectorized access to random numbers and a few
            other utilities.

        """
        return self._manager.get_randomness_stream(decision_point, for_initialization)

    def get_seed(self, decision_point: str) -> int:
        """Get a randomly generated seed for use with external randomness tools.

        Parameters
        ----------
        decision_point :
            A unique identifier for a stream of random numbers.  Typically
            represents a decision that needs to be made each time step like
            'moves_left' or 'gets_disease'.

        Returns
        -------
        int
            A seed for a random number generation that is linked to Vivarium's
            common random number framework.

        """
        return self._manager.get_seed(decision_point)

    def register_simulants(self, simulants: pd.DataFrame):
        """Registers simulants with the Common Random Number Framework.

        Parameters
        ----------
        simulants
            A section of the state table with new simulants and at least the
            columns specified in
            ``builder.configuration.randomness.key_columns``.  This function
            should be called as soon as the key columns are generated.

        """
        self._manager.register_simulants(simulants)



import matplotlib.pyplot as plt


def plot_birds(simulation, plot_velocity=False):
    width = simulation.configuration.location.width
    height = simulation.configuration.location.height
    pop = simulation.get_population()

    plt.figure(figsize=[12, 12])
    plt.scatter(pop.x, pop.y, color=pop.color)
    if plot_velocity:
        plt.quiver(pop.x, pop.y, pop.vx, pop.vy, color=pop.color, width=0.002)
    plt.xlabel('x')
    plt.ylabel('y')
    plt.axis([0, width, 0, height])
    plt.show()



import pandas as pd
import pytest

from vivarium.framework.population import PopulationView, InitializerComponentSet, PopulationError, PopulationManager


class DummyPopulationManager:
    def __init__(self):
        self.get_population = lambda _ : pd.DataFrame({'age': [0, 10, 20, 30, 40, 50, 60, 70], 'sex': ['Male', 'Female']*4})


def test_create_PopulationView_with_all_columns():
    manager = DummyPopulationManager()
    view = PopulationView(manager, 0)
    assert set(view.columns) == {'age', 'sex'}


def test_initializer_set_fail_type():
    component_set = InitializerComponentSet()

    with pytest.raises(TypeError):
        component_set.add(lambda: 'test', ['test_column'])

    def initializer():
        return 'test'

    with pytest.raises(TypeError):
        component_set.add(initializer, ['test_column'])


class UnnamedComponent:

    def initializer(self):
        return 'test'


class Component:

    def __init__(self, name):
        self.name = name

    def initializer(self):
        return 'test'

    def other_initializer(self):
        return 'whoops'


def test_initializer_set_fail_attr():
    component_set = InitializerComponentSet()

    with pytest.raises(AttributeError):
        component_set.add(UnnamedComponent().initializer, ['test_column'])


def test_initializer_set_duplicate_component():
    component_set = InitializerComponentSet()
    component = Component('test')

    component_set.add(component.initializer, ['test_column1'])
    with pytest.raises(PopulationError, match='multiple population initializers'):
        component_set.add(component.other_initializer, ['test_column2'])


def test_initializer_set_duplicate_columns():
    component_set = InitializerComponentSet()
    component1 = Component('test1')
    component2 = Component('test2')
    columns = ['test_column']

    component_set.add(component1.initializer, columns)
    with pytest.raises(PopulationError, match='both registered initializers'):
        component_set.add(component2.initializer, columns)

    with pytest.raises(PopulationError, match='both registered initializers'):
        component_set.add(component2.initializer, ['sneaky_column'] + columns)


def test_initializer_set():
    component_set = InitializerComponentSet()
    for i in range(10):
        component = Component(i)
        columns = [f'test_column_{i}_{j}' for j in range(5)]
        component_set.add(component.initializer, columns)


def test_get_view_with_no_query():
    manager = PopulationManager()
    view = manager._get_view(columns=['age','sex'])
    assert view.query == 'tracked == True'



import sys
if sys.version_info < (3, 6) or sys.version_info >= (3, 9):
    # Python 3.5 does not support f-strings
    py_version = '.'.join([str(v) for v in sys.version_info[:3]])
    error = ('\n----------------------------------------\n'
            'Error: Vivarium runs under python 3.6-3.8.\n'
            'You are running python {py_version}'.format(py_version = py_version))
    print(error, file=sys.stderr)
    sys.exit(1)


from pathlib import Path

from setuptools import setup, find_packages


if __name__ == "__main__":

    base_dir = Path(__file__).parent
    src_dir = base_dir / 'src'

    about = {}
    with (src_dir / "vivarium" / "__about__.py").open() as f:
        exec(f.read(), about)

    with (base_dir / "README.rst").open() as f:
        long_description = f.read()

    install_requirements = [
        'numpy',
        'pandas',
        'pyyaml>=5.1',
        'scipy',
        'click',
        'tables',
        'networkx',
        'loguru',
    ]

    interactive_requirements = [
        'IPython',
        'ipywidgets',
        'jupyter',
    ]

    test_requirements = [
        'pytest',
        'pytest-mock',
    ]

    doc_requirements = [
        'sphinx>=4.0',
        'sphinx-rtd-theme',
        'sphinx-click',
        'IPython',
        'matplotlib'
    ]

    setup(
        name=about['__title__'],
        version=about['__version__'],

        description=about['__summary__'],
        long_description=long_description,
        license=about['__license__'],
        url=about["__uri__"],

        author=about["__author__"],
        author_email=about["__email__"],

        classifiers=[
            "Intended Audience :: Developers",
            "Intended Audience :: Education",
            "Intended Audience :: Science/Research",
            "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
            "Natural Language :: English",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: POSIX",
            "Operating System :: POSIX :: BSD",
            "Operating System :: POSIX :: Linux",
            "Operating System :: Microsoft :: Windows",
            "Programming Language :: Python",
            "Programming Language :: Python :: Implementation :: CPython",
            "Topic :: Education",
            "Topic :: Scientific/Engineering",
            "Topic :: Scientific/Engineering :: Artificial Life",
            "Topic :: Scientific/Engineering :: Mathematics",
            "Topic :: Scientific/Engineering :: Medical Science Apps.",
            "Topic :: Scientific/Engineering :: Physics",
            "Topic :: Software Development :: Libraries",
        ],

        package_dir={'': 'src'},
        packages=find_packages(where='src'),
        include_package_data=True,

        install_requires=install_requirements,
        tests_require=test_requirements,
        extras_require={
            'docs': doc_requirements,
            'test': test_requirements,
            'interactive': interactive_requirements,
            'dev': doc_requirements + test_requirements + interactive_requirements,
        },

        entry_points="""
                [console_scripts]
                simulate=vivarium.interface.cli:simulate
            """,

        zip_safe=False,
    )



"""
===================
The Vivarium Engine
===================

The engine houses the :class:`SimulationContext` -- the key ``vivarium`` object
for running and interacting with simulations. It is the top level manager
for all state information in ``vivarium``.  By intention, it exposes a very
simple interface for managing the
:ref:`simulation lifecycle <lifecycle_concept>`.

Also included here is the simulation :class:`Builder`, which is the main
interface that components use to interact with the simulation framework. You
can read more about how the builder works and what services is exposes
:ref:`here <builder_concept>`.

Finally, there are a handful of wrapper methods that allow a user or user
tools to easily setup and run a simulation.

"""
from pathlib import Path
from pprint import pformat
from typing import Union, List, Dict

from loguru import logger

from vivarium.config_tree import ConfigTree
from vivarium.framework.configuration import build_model_specification
from .artifact import ArtifactInterface
from .components import ComponentInterface
from .event import EventInterface
from .lookup import LookupTableInterface
from .metrics import Metrics
from .plugins import PluginManager
from .population import PopulationInterface
from .resource import ResourceInterface
from .randomness import RandomnessInterface
from .values import ValuesInterface
from .time import TimeInterface
from .lifecycle import LifeCycleInterface
from .results import ResultsInterface


class SimulationContext:

    def __init__(self, model_specification: Union[str, Path, ConfigTree] = None,
                 components: Union[List, Dict, ConfigTree] = None,
                 configuration: Union[Dict, ConfigTree] = None,
                 plugin_configuration: Union[Dict, ConfigTree] = None):
        # Bootstrap phase: Parse arguments, make private managers
        component_configuration = components if isinstance(components, (dict, ConfigTree)) else None
        self._additional_components = components if isinstance(components, List) else []
        model_specification = build_model_specification(model_specification, component_configuration,
                                                        configuration, plugin_configuration)

        self._plugin_configuration = model_specification.plugins
        self._component_configuration = model_specification.components
        self.configuration = model_specification.configuration

        self._plugin_manager = PluginManager(model_specification.plugins)

        # TODO: Setup logger here.

        self._builder = Builder(self.configuration, self._plugin_manager)

        # This formally starts the initialization phase (this call makes the
        # life-cycle manager).
        self._lifecycle = self._plugin_manager.get_plugin('lifecycle')
        self._lifecycle.add_phase('setup', ['setup', 'post_setup', 'population_creation'])
        self._lifecycle.add_phase(
            'main_loop', ['time_step__prepare', 'time_step', 'time_step__cleanup', 'collect_metrics'], loop=True
        )
        self._lifecycle.add_phase('simulation_end', ['simulation_end', 'report'])

        self._component_manager = self._plugin_manager.get_plugin('component_manager')
        self._component_manager.setup(self.configuration, self._lifecycle)

        self._clock = self._plugin_manager.get_plugin('clock')
        self._values = self._plugin_manager.get_plugin('value')
        self._events = self._plugin_manager.get_plugin('event')
        self._population = self._plugin_manager.get_plugin('population')
        self._resource = self._plugin_manager.get_plugin('resource')
        self._results = self._plugin_manager.get_plugin('results')
        self._tables = self._plugin_manager.get_plugin('lookup')
        self._randomness = self._plugin_manager.get_plugin('randomness')
        self._data = self._plugin_manager.get_plugin('data')

        for name, controller in self._plugin_manager.get_optional_controllers().items():
            setattr(self, f'_{name}', controller)

        # The order the managers are added is important.  It represents the
        # order in which they will be set up.  The clock is required by
        # several of the other managers, including the lifecycle manager.  The
        # lifecycle manager is also required by most managers. The randomness
        # manager requires the population manager.  The remaining managers need
        # no ordering.
        managers = [self._clock, self._lifecycle, self._resource, self._values,
                    self._population, self._randomness, self._events, self._tables,
                    self._data, self._results] + list(self._plugin_manager.get_optional_controllers().values())
        self._component_manager.add_managers(managers)

        component_config_parser = self._plugin_manager.get_plugin('component_configuration_parser')
        # Tack extra components onto the end of the list generated from the model specification.
        components = (component_config_parser.get_components(self._component_configuration)
                      + self._additional_components
                      + [Metrics()])

        self._lifecycle.add_constraint(self.add_components, allow_during=['initialization'])
        self._lifecycle.add_constraint(self.get_population, restrict_during=['initialization', 'setup', 'post_setup'])

        self.add_components(components)

    @property
    def name(self):
        return 'simulation_context'

    def setup(self):
        self._lifecycle.set_state('setup')
        self.configuration.freeze()
        self._component_manager.setup_components(self._builder)

        self.simulant_creator = self._builder.population.get_simulant_creator()

        self.time_step_events = self._lifecycle.get_state_names('main_loop')
        self.time_step_emitters = {k: self._builder.event.get_emitter(k) for k in self.time_step_events}
        self.end_emitter = self._builder.event.get_emitter('simulation_end')

        post_setup = self._builder.event.get_emitter('post_setup')
        self._lifecycle.set_state('post_setup')
        post_setup(None)

    def initialize_simulants(self):
        self._lifecycle.set_state('population_creation')
        pop_params = self.configuration.population
        # Fencepost the creation of the initial population.
        self._clock.step_backward()
        population_size = pop_params.population_size
        self.simulant_creator(population_size, population_configuration={'sim_state': 'setup'})
        self._clock.step_forward()

    def step(self):
        logger.debug(self._clock.time)
        for event in self.time_step_events:
            self._lifecycle.set_state(event)
            self.time_step_emitters[event](self._population.get_population(True).index)
        self._clock.step_forward()

    def run(self):
        while self._clock.time < self._clock.stop_time:
            self.step()

    def finalize(self):
        self._lifecycle.set_state('simulation_end')
        self.end_emitter(self._population.get_population(True).index)
        unused_config_keys = self.configuration.unused_keys()
        if unused_config_keys:
            logger.debug(f"Some configuration keys not used during run: {unused_config_keys}.")

    def report(self, print_results=True):
        self._lifecycle.set_state('report')
        metrics = self._values.get_value('metrics')(self._population.get_population(True).index)
        if print_results:
            logger.debug(pformat(metrics))
        return metrics

    def add_components(self, component_list):
        """Adds new components to the simulation."""
        self._component_manager.add_components(component_list)

    def get_population(self, untracked: bool = True):
        return self._population.get_population(untracked)

    def __repr__(self):
        return "SimulationContext()"

    def __str__(self):
        return str(self._lifecycle)


class Builder:
    """Toolbox for constructing and configuring simulation components.

    This is the access point for components through which they are able to
    utilize a variety of interfaces to interact with the simulation framework.

    Attributes
    ----------
    lookup: LookupTableInterface
        Provides access to simulant-specific data via the
        :ref:`lookup table<lookup_concept>` abstraction.
    value: ValuesInterface
        Provides access to computed simulant attribute values via the
        :ref:`value pipeline<values_concept>` system.
    event: EventInterface
        Provides access to event listeners utilized in the
        :ref:`event<event_concept>` system.
    population: PopulationInterface
        Provides access to simulant state table via the
        :ref:`population<population_concept>` system.
    resource: ResourceInterface
        Provides access to the :ref:`resource<resource_concept>` system,
        which manages dependencies between components.
    time: TimeInterface
        Provides access to the simulation's :ref:`clock<time_concept>`.
    components: ComponentInterface
        Provides access to the :ref:`component management<components_concept>`
        system, which maintains a reference to all managers and components in
        the simulation.
    lifecycle: LifeCycleInterface
        Provides access to the :ref:`life-cycle<lifecycle_concept>` system,
        which manages the simulation's execution life-cycle.
    data: ArtifactInterface
        Provides access to the simulation's input data housed in the
        :ref:`data artifact<data_concept>`.

    Notes
    -----
    A `Builder` should never be created directly. It will automatically be
    created during the initialization of a :class:`SimulationContext`

    """

    def __init__(self, configuration, plugin_manager):
        self.configuration = configuration

        self.lookup = plugin_manager.get_plugin_interface('lookup')                 # type: LookupTableInterface
        self.value = plugin_manager.get_plugin_interface('value')                   # type: ValuesInterface
        self.event = plugin_manager.get_plugin_interface('event')                   # type: EventInterface
        self.population = plugin_manager.get_plugin_interface('population')         # type: PopulationInterface
        self.resources = plugin_manager.get_plugin_interface('resource')            # type: ResourceInterface
        self.results = plugin_manager.get_plugin_interface('results')               # type: ResultsInterface
        self.randomness = plugin_manager.get_plugin_interface('randomness')         # type: RandomnessInterface
        self.time = plugin_manager.get_plugin_interface('clock')                    # type: TimeInterface
        self.components = plugin_manager.get_plugin_interface('component_manager')  # type: ComponentInterface
        self.lifecycle = plugin_manager.get_plugin_interface('lifecycle')           # type: LifeCycleInterface
        self.data = plugin_manager.get_plugin_interface('data')                     # type: ArtifactInterface


        for name, interface in plugin_manager.get_optional_interfaces().items():
            setattr(self, name, interface)

    def __repr__(self):
        return "Builder()"


def run_simulation(model_specification: Union[str, Path, ConfigTree] = None,
                   components: Union[List, Dict, ConfigTree] = None,
                   configuration: Union[Dict, ConfigTree] = None,
                   plugin_configuration: Union[Dict, ConfigTree] = None):
    simulation = SimulationContext(model_specification, components, configuration, plugin_configuration)
    simulation.setup()
    simulation.initialize_simulants()
    simulation.run()
    simulation.finalize()
    return simulation



import numpy as np
import pandas as pd

from vivarium.framework.engine import Builder
from vivarium.framework.event import Event


class Mortality:
    """Introduces death into the simulation.

    Attributes
    ----------
    configuration_defaults :
        A set of default configuration values for this component. These can be
        overwritten in the simulation model specification or by providing
        override values when constructing an interactive simulation.
    """

    configuration_defaults = {
        'mortality': {
            'mortality_rate': 0.01,
        }
    }

    def __init__(self):
        self.name = 'mortality'

    # noinspection PyAttributeOutsideInit
    def setup(self, builder: Builder):
        """Performs this component's simulation setup.

        The ``setup`` method is automatically called by the simulation
        framework. The framework passes in a ``builder`` object which
        provides access to a variety of framework subsystems and metadata.

        Parameters
        ----------
        builder :
            Access to simulation tools and subsystems.
        """
        self.config = builder.configuration.mortality
        self.population_view = builder.population.get_view(['alive'], query="alive == 'alive'")
        self.randomness = builder.randomness.get_stream('mortality')

        self.mortality_rate = builder.value.register_rate_producer('mortality_rate', source=self.base_mortality_rate)

        builder.event.register_listener('time_step', self.determine_deaths)

    def base_mortality_rate(self, index: pd.Index) -> pd.Series:
        """Computes the base mortality rate for every individual.

        Parameters
        ----------
        index :
            A representation of the simulants to compute the base mortality
            rate for.

        Returns
        -------
            The base mortality rate for all simulants in the index.
        """
        return pd.Series(self.config.mortality_rate, index=index)

    def determine_deaths(self, event: Event):
        """Determines who dies each time step.

        Parameters
        ----------
        event :
            An event object emitted by the simulation containing an index
            representing the simulants affected by the event and timing
            information.
        """
        effective_rate = self.mortality_rate(event.index)
        effective_probability = 1 - np.exp(-effective_rate)
        draw = self.randomness.get_draw(event.index)
        affected_simulants = draw < effective_probability
        self.population_view.update(pd.Series('dead', index=event.index[affected_simulants]))



"""
================================
The Population Management System
================================

This module provides tools for managing the :term:`state table <State Table>`
in a :mod:`vivarium` simulation, which is the record of all simulants in a
simulation and their state. It's main tasks are managing the creation of new
simulants and providing the ability for components to view and update simulant
state safely during runtime.

"""
from types import MethodType
from typing import List, Callable, Union, Dict, Any, NamedTuple, Tuple

import pandas as pd

from vivarium.exceptions import VivariumError


class PopulationError(VivariumError):
    """Error raised when the population is invalidly queried or updated."""
    pass


class PopulationView:
    """A read/write manager for the simulation state table.

    It can be used to both read and update the state of the population. A
    PopulationView can only read and write columns for which it is configured.
    Attempts to update non-existent columns are ignored except during
    simulant creation when new columns are allowed to be created.

    Parameters
    ----------
    manager
        The population manager for the simulation.
    columns
        The set of columns this view should have access too.  If explicitly
        specified as ``None``, this view will have access to the entire
        state table.
    query
        A :mod:`pandas`-style filter that will be applied any time this
        view is read from.

    Notes
    -----
    By default, this view will filter out ``untracked`` simulants unless
    the ``tracked`` column is specified in the initialization arguments.

    """

    def __init__(self,
                 manager: 'PopulationManager',
                 view_id: int,
                 columns: Union[List[str], Tuple[str], None] = (),
                 query: str = None):
        self._manager = manager
        self._id = view_id
        self._columns = list(columns)
        self._query = query

    @property
    def name(self):
        return f'population_view_{self._id}'

    @property
    def columns(self) -> List[str]:
        """The columns that the view can read and update.

        If the view was created with ``None`` as the columns argument, then
        the view will have access to the full table by default. That case
        should be only be used in situations where the full state table is
        actually needed, like for some metrics collection applications.

        """
        if not self._columns:
            return list(self._manager.get_population(True).columns)
        return list(self._columns)

    @property
    def query(self) -> str:
        """A :mod:`pandas` style query to filter the rows of this view.

        This query will be applied any time the view is read. This query may
        reference columns not in the view's columns.

        """
        return self._query

    def subview(self, columns: Union[List[str], Tuple[str]]) -> 'PopulationView':
        """Retrieves a new view with a subset of this view's columns.

        Parameters
        ----------
        columns
            The set of columns to provide access to in the subview. Must be
            a proper subset of this view's columns.

        Returns
        -------
        PopulationView
            A new view with access to the requested columns.

        Raises
        ------
        PopulationError
            If the requested columns are not a proper subset of this view's
            columns.

        Notes
        -----
        Subviews are useful during population initialization. The original
        view may contain both columns that a component needs to create and
        update as well as columns that the component needs to read.  By
        requesting a subview, a component can read the sections it needs
        without running the risk of trying to access uncreated columns
        because the component itself has not created them.

        """
        if set(columns) > set(self.columns):
            raise PopulationError(f"Invalid subview requested.  Requested columns must be a subset of this "
                                  f"view's columns.  Requested columns: {columns}, Available columns: {self.columns}")
        # Skip constraints for requesting subviews.
        return self._manager._get_view(columns, self.query)

    def get(self, index: pd.Index, query: str = '') -> pd.DataFrame:
        """Select the rows represented by the given index from this view.

        For the rows in ``index`` get the columns from the simulation's
        state table to which this view has access. The resulting rows may be
        further filtered by the view's query and only return a subset
        of the population represented by the index.

        Parameters
        ----------
        index
            Index of the population to get.
        query
            Additional conditions used to filter the index. These conditions
            will be unioned with the default query of this view.  The query
            provided may use columns that this view does not have access to.

        Returns
        -------
        pandas.DataFrame
            A table with the subset of the population requested.

        Raises
        ------
        PopulationError
            If this view has access to columns that have not yet been created
            and this method is called.  If you see this error, you should
            request a subview with the columns you need read access to.

        See Also
        --------
        :meth:`subview <PopulationView.subview>`

        """
        pop = self._manager.get_population(True).loc[index]

        if not index.empty:
            if self._query:
                pop = pop.query(self._query)
            if query:
                pop = pop.query(query)

        if not self._columns:
            return pop
        else:
            columns = self._columns
            non_existent_columns = set(columns) - set(pop.columns)
            if non_existent_columns:
                raise PopulationError(f'Requested column(s) {non_existent_columns} not in population table.')
            else:
                return pop.loc[:, columns]

    def update(self, population_update: Union[pd.DataFrame, pd.Series]):
        """Updates the state table with the provided data.

        Parameters
        ----------
        population_update
            The data which should be copied into the simulation's state. If
            the update is a :class:`pandas.DataFrame`, it can contain a subset
            of the view's columns but no extra columns. If ``pop`` is a
            :class:`pandas.Series` it must have a name that matches one of
            this view's columns unless the view only has one column in which
            case the Series will be assumed to refer to that regardless of its
            name.

        Raises
        ------
        PopulationError
            If the provided data name or columns does not match columns that
            this view manages or if the view is being updated with a data
            type inconsistent with the original population data.

        """
        if population_update.empty:
            return

        # TODO: Cast series to data frame and clean this up.
        if isinstance(population_update, pd.Series):
            if population_update.name in self._columns:
                affected_columns = [population_update.name]
            elif len(self._columns) == 1:
                affected_columns = self._columns
            else:
                raise PopulationError('Cannot update with a pandas series unless the series name is a column '
                                      'name in the view or there is only a single column in the view.')
        else:
            if not set(population_update.columns).issubset(self._columns):
                raise PopulationError(f'Cannot update with a DataFrame that contains columns the view does not. '
                                      f'Dataframe contains the following extra columns: '
                                      f'{set(population_update.columns).difference(self._columns)}.')
            affected_columns = set(population_update.columns)

        affected_columns = set(affected_columns).intersection(self._columns)
        state_table = self._manager.get_population(True)
        if not self._manager.growing:
            affected_columns = set(affected_columns).intersection(state_table.columns)

        for affected_column in affected_columns:
            if affected_column in state_table:
                new_state_table_values = state_table[affected_column].values
                if isinstance(population_update, pd.Series):
                    update_values = population_update.values
                else:
                    update_values = population_update[affected_column].values
                new_state_table_values[population_update.index] = update_values

                if new_state_table_values.dtype != update_values.dtype:
                    # This happens when the population is being grown because extending
                    # the index forces columns that don't have a natural null type
                    # to become 'object'
                    if not self._manager.growing:
                        raise PopulationError('Component corrupting population table. '
                                              f'Column name: {affected_column} '
                                              f'Old column type: {new_state_table_values.dtype} '
                                              f'New column type: {update_values.dtype}')
                    new_state_table_values = new_state_table_values.astype(update_values.dtype)
            else:
                if isinstance(population_update, pd.Series):
                    new_state_table_values = population_update.values
                else:
                    new_state_table_values = population_update[affected_column].values
            self._manager._population[affected_column] = new_state_table_values

    def __repr__(self):
        return f"PopulationView(_id={self._id}, _columns={self.columns}, _query={self._query})"


class SimulantData(NamedTuple):
    """Data to help components initialize simulants.

    Any time simulants are added to the simulation, each initializer is called
    with this structure containing information relevant to their
    initialization.

    """
    #: The index representing the new simulants being added to the simulation.
    index: pd.Index
    #: A dictionary of extra data passed in by the component creating the
    #: population.
    user_data: Dict[str, Any]
    #: The time when the simulants enter the simulation.
    creation_time: pd.Timestamp
    #: The span of time over which the simulants are created.  Useful for,
    #: e.g., distributing ages over the window.
    creation_window: pd.Timedelta


class InitializerComponentSet:
    """Set of unique components with population initializers."""

    def __init__(self):
        self._components = {}
        self._columns_produced = {}

    def add(self, initializer: Callable, columns_produced: List[str]):
        """Adds an initializer and columns to the set, enforcing uniqueness.

        Parameters
        ----------
        initializer
            The population initializer to add to the set.
        columns_produced
            The columns the initializer produces.

        Raises
        ------
        TypeError
            If the initializer is not an object method.
        AttributeError
            If the object bound to the method does not have a name attribute.
        PopulationError
            If the component bound to the method already has an initializer
            registered or if the columns produced are duplicates of columns
            another initializer produces.

        """
        if not isinstance(initializer, MethodType):
            raise TypeError('Population initializers must be methods of named simulation components. '
                            f'You provided {initializer} which is of type {type(initializer)}.')
        component = initializer.__self__
        if not hasattr(component, "name"):
            raise AttributeError('Population initializers must be methods of named simulation components. '
                                 f'You provided {initializer} which is bound to {component} that has no '
                                 f'name attribute.')
        if component.name in self._components:
            raise PopulationError(f'Component {component.name} has multiple population initializers. '
                                  'This is not allowed.')
        for column in columns_produced:
            if column in self._columns_produced:
                raise PopulationError(f'Component {component.name} and component {self._columns_produced[column]} '
                                      f'have both registered initializers for column {column}.')
            self._columns_produced[column] = component.name
        self._components[component.name] = columns_produced

    def __repr__(self):
        return repr(self._components)

    def __str__(self):
        return str(self._components)


class PopulationManager:
    """Manages the state of the simulated population."""

    # TODO: Move the configuration for initial population creation to
    # user components.
    configuration_defaults = {
        'population': {'population_size': 100}
    }

    def __init__(self):
        self._population = pd.DataFrame()
        self._initializer_components = InitializerComponentSet()
        self.growing = False
        self._last_id = -1

    ############################
    # Normal Component Methods #
    ############################

    @property
    def name(self):
        """The name of this component."""
        return "population_manager"

    def setup(self, builder):
        """Registers the population manager with other vivarium systems."""
        self.clock = builder.time.clock()
        self.step_size = builder.time.step_size()
        self.resources = builder.resources
        self._add_constraint = builder.lifecycle.add_constraint

        builder.lifecycle.add_constraint(self.get_view, allow_during=['setup', 'post_setup', 'population_creation',
                                                                      'simulation_end', 'report'])
        builder.lifecycle.add_constraint(self.get_simulant_creator, allow_during=['setup'])
        builder.lifecycle.add_constraint(self.register_simulant_initializer, allow_during=['setup'])

        self.register_simulant_initializer(self.on_initialize_simulants, creates_columns=['tracked'])
        self._view = self.get_view(['tracked'])

        builder.value.register_value_modifier('metrics', modifier=self.metrics)

    def on_initialize_simulants(self, pop_data: SimulantData):
        """Adds a ``tracked`` column to the state table for new simulants."""
        status = pd.Series(True, index=pop_data.index)
        self._view.update(status)

    def metrics(self, index, metrics):
        """Reports tracked and untracked population sizes at simulation end."""
        population = self._view.get(index)
        untracked = population[~population.tracked]
        tracked = population[population.tracked]

        metrics['total_population_untracked'] = len(untracked)
        metrics['total_population_tracked'] = len(tracked)
        metrics['total_population'] = len(untracked)+len(tracked)
        return metrics

    def __repr__(self):
        return "PopulationManager()"

    ###########################
    # Builder API and helpers #
    ###########################

    def get_view(self, columns: Union[List[str], Tuple[str]], query: str = None) -> PopulationView:
        """Get a time-varying view of the population state table.

        The requested population view can be used to view the current state or
        to update the state with new values.

        If the column 'tracked' is not specified in the ``columns`` argument,
        the query string 'tracked == True' will be added to the provided
        query argument. This allows components to ignore untracked simulants
        by default.

        Parameters
        ----------
        columns
            A subset of the state table columns that will be available in the
            returned view.
        query
            A filter on the population state.  This filters out particular
            simulants (rows in the state table) based on their current state.
            The query should be provided in a way that is understood by the
            :meth:`pandas.DataFrame.query` method and may reference state
            table columns not requested in the ``columns`` argument.

        Returns
        -------
        PopulationView
            A filtered view of the requested columns of the population state
            table.

        """
        view = self._get_view(columns, query)
        self._add_constraint(view.get, restrict_during=['initialization', 'setup', 'post_setup'])
        self._add_constraint(view.update, restrict_during=['initialization', 'setup', 'post_setup',
                                                           'simulation_end', 'report'])
        return view

    def _get_view(self, columns: Union[List[str], Tuple[str]], query: str = None):
        if columns and 'tracked' not in columns:
            if query is None:
                query = 'tracked == True'
            elif 'tracked' not in query:
                query += 'and tracked == True'
        self._last_id += 1
        return PopulationView(self, self._last_id, columns, query)

    def register_simulant_initializer(self, initializer: Callable,
                                      creates_columns: List[str] = (),
                                      requires_columns: List[str] = (),
                                      requires_values: List[str] = (),
                                      requires_streams: List[str] = ()):
        """Marks a source of initial state information for new simulants.

        Parameters
        ----------
        initializer
            A callable that adds or updates initial state information about
            new simulants.
        creates_columns
            A list of the state table columns that the given initializer
            provides the initial state information for.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the provided initializer
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the provided initializer is called.
        requires_streams
            A list of the randomness streams necessary to initialize the
            simulant attributes.

        """
        self._initializer_components.add(initializer, creates_columns)
        dependencies = ([f'column.{name}' for name in requires_columns]
                        + [f'value.{name}' for name in requires_values]
                        + [f'stream.{name}' for name in requires_streams])
        if creates_columns != ['tracked']:
            # The population view itself uses the tracked column, so include
            # to be safe.
            dependencies += ['column.tracked']
        self.resources.add_resources('column', list(creates_columns), initializer, dependencies)

    def get_simulant_creator(self) -> Callable:
        """Gets a function that can generate new simulants.

        Returns
        -------
           The simulant creator function. The creator function takes the
           number of simulants to be created as it's first argument and a dict
           population configuration that will be available to simulant
           initializers as it's second argument. It generates the new rows in
           the population state table and then calls each initializer
           registered with the population system with a data
           object containing the state table index of the new simulants, the
           configuration info passed to the creator, the current simulation
           time, and the size of the next time step.

        """
        return self._create_simulants

    def _create_simulants(self, count: int, population_configuration: Dict[str, Any] = None) -> pd.Index:
        population_configuration = population_configuration if population_configuration else {}
        new_index = range(len(self._population) + count)
        new_population = self._population.reindex(new_index)
        index = new_population.index.difference(self._population.index)
        self._population = new_population
        self.growing = True
        for initializer in self.resources:
            initializer(SimulantData(index, population_configuration, self.clock(), self.step_size()))
        self.growing = False
        return index

    ###############
    # Context API #
    ###############

    def get_population(self, untracked: bool) -> pd.DataFrame:
        """Provides a copy of the full population state table.

        Parameters
        ----------
        untracked
            Whether to include untracked simulants in the returned population.

        Returns
        -------
        pandas.DataFrame
            A copy of the population table.

        """
        pop = self._population.copy()
        if not untracked:
            pop = pop[pop.tracked]
        return pop


class PopulationInterface:
    """Provides access to the system for reading and updating the population.

    The most important aspect of the simulation state is the ``population
    table`` or ``state table``.  It is a table with a row for every
    individual or cohort (referred to as a simulant) being simulated and a
    column for each of the attributes of the simulant being modeled.  All
    access to the state table is mediated by
    :class:`population views <PopulationView>`, which may be requested from
    this system during setup time.

    The population system itself manages a single attribute of simulants
    called ``tracked``. This attribute allows global control of which
    simulants are available to read and update in the state table by
    default.

    For example, in a simulation of childhood illness, we might not
    need information about individuals or cohorts once they reach five years
    of age, and so we can have them "age out" of the simulation at five years
    old by setting the ``tracked`` attribute to ``False``.

    """

    def __init__(self, manager: PopulationManager):
        self._manager = manager

    def get_view(self, columns: Union[List[str], Tuple[str]], query: str = None) -> PopulationView:
        """Get a time-varying view of the population state table.

        The requested population view can be used to view the current state or
        to update the state with new values.

        If the column 'tracked' is not specified in the ``columns`` argument,
        the query string 'tracked == True' will be added to the provided
        query argument. This allows components to ignore untracked simulants
        by default.

        Parameters
        ----------
        columns
            A subset of the state table columns that will be available in the
            returned view.
        query
            A filter on the population state.  This filters out particular
            simulants (rows in the state table) based on their current state.
            The query should be provided in a way that is understood by the
            :meth:`pandas.DataFrame.query` method and may reference state
            table columns not requested in the ``columns`` argument.

        Returns
        -------
        PopulationView
            A filtered view of the requested columns of the population state
            table.

        """
        return self._manager.get_view(columns, query)

    def get_simulant_creator(self) -> Callable[[int, Union[Dict[str, Any], None]], pd.Index]:
        """Gets a function that can generate new simulants.

        Returns
        -------
           The simulant creator function. The creator function takes the
           number of simulants to be created as it's first argument and a dict
           population configuration that will be available to simulant
           initializers as it's second argument. It generates the new rows in
           the population state table and then calls each initializer
           registered with the population system with a data
           object containing the state table index of the new simulants, the
           configuration info passed to the creator, the current simulation
           time, and the size of the next time step.

        """
        return self._manager.get_simulant_creator()

    def initializes_simulants(self, initializer: Callable[[SimulantData], None],
                              creates_columns: List[str] = (),
                              requires_columns: List[str] = (),
                              requires_values: List[str] = (),
                              requires_streams: List[str] = ()):
        """Marks a source of initial state information for new simulants.

        Parameters
        ----------
        initializer
            A callable that adds or updates initial state information about
            new simulants.
        creates_columns
            A list of the state table columns that the given initializer
            provides the initial state information for.
        requires_columns
            A list of the state table columns that already need to be present
            and populated in the state table before the provided initializer
            is called.
        requires_values
            A list of the value pipelines that need to be properly sourced
            before the provided initializer is called.
        requires_streams
            A list of the randomness streams necessary to initialize the
            simulant attributes.

        """
        self._manager.register_simulant_initializer(initializer, creates_columns,
                                                    requires_columns, requires_values, requires_streams)



"""
=======================
Configuration Utilities
=======================

A set of functions for turning model specification files into programmatic
representations of :term:`model specifications <Model Specification>` and
:term:`configurations <Configuration>`.

"""
from pathlib import Path
from typing import Union, Dict

import yaml

from vivarium.config_tree import ConfigTree, ConfigurationError
from vivarium.framework.plugins import DEFAULT_PLUGINS


def build_model_specification(model_specification: Union[str, Path, ConfigTree] = None,
                              component_configuration: Union[Dict, ConfigTree] = None,
                              configuration: Union[Dict, ConfigTree] = None,
                              plugin_configuration: Union[Dict, ConfigTree] = None) -> ConfigTree:
    if isinstance(model_specification, (str, Path)):
        validate_model_specification_file(model_specification)
        source = str(model_specification)
    else:
        source = 'user_supplied_args'

    output_spec = _get_default_specification()
    output_spec.update(model_specification, layer='model_override', source=source)
    output_spec.components.update(component_configuration, layer='override', source='user_supplied_args')
    output_spec.configuration.update(configuration, layer='override', source='user_supplied_args')
    output_spec.plugins.update(plugin_configuration, layer='override', source='user_supplied_args')

    return output_spec


def validate_model_specification_file(file_path: Union[str, Path]) -> None:
    """Ensures the provided file is a yaml file"""
    file_path = Path(file_path)
    if not file_path.exists():
        raise ConfigurationError('If you provide a model specification file, it must be a file. '
                                 f'You provided {str(file_path)}', value_name=None)

    if file_path.suffix not in ['.yaml', '.yml']:
        raise ConfigurationError(f'Model specification files must be in a yaml format. You provided {file_path.suffix}',
                                 value_name=None)
    # Attempt to load
    with file_path.open() as f:
        raw_spec = yaml.full_load(f)
    top_keys = set(raw_spec.keys())
    valid_keys = {'plugins', 'components', 'configuration'}
    if not top_keys <= valid_keys:
        raise ConfigurationError(f'Model specification contains additional top level '
                                 f'keys {valid_keys.difference(top_keys)}.', value_name=None)


def build_simulation_configuration() -> ConfigTree:
    return _get_default_specification().configuration


def _get_default_specification():
    default_config_layers = ['base', 'user_configs', 'component_configs', 'model_override', 'override']
    default_metadata = {'layer': 'base', 'source': 'vivarium_defaults'}

    model_specification = ConfigTree(layers=default_config_layers)
    model_specification.update(DEFAULT_PLUGINS, **default_metadata)
    model_specification.update({'components': {},
                                'configuration': {}})

    user_config_path = Path('~/vivarium.yaml').expanduser()
    if user_config_path.exists():
        model_specification.configuration.update(user_config_path, layer='user_configs')

    return model_specification



import networkx as nx
import pytest

from vivarium.framework.resource import (ResourceGroup, ResourceManager,
                                         RESOURCE_TYPES, NULL_RESOURCE_TYPE, ResourceError)


class Component:
    def __init__(self, name):
        self.name = name

    def producer(self):
        return 'resources!'


def test_resource_group():
    c = Component('base')
    r_type = 'column'
    r_names = [str(i) for i in range(5)]
    r_producer = c.producer
    r_dependencies = []

    rg = ResourceGroup(r_type, r_names, r_producer, r_dependencies)

    assert rg.type == r_type
    assert rg.names == [f'{r_type}.{name}' for name in r_names]
    assert rg.producer == c.producer
    assert not rg.dependencies
    assert list(rg) == rg.names


def test_resource_manager_get_resource_group():
    rm = ResourceManager()
    c = Component('base')
    r_type = 'column'
    r_names = [str(i) for i in range(5)]
    r_producer = c.producer
    r_dependencies = []

    rg = rm._get_resource_group(r_type, r_names, r_producer, r_dependencies)

    assert rg.type == r_type
    assert rg.names == [f'{r_type}.{name}' for name in r_names]
    assert rg.producer == c.producer
    assert not rg.dependencies
    assert list(rg) == rg.names


def test_resource_manager_get_resource_group_null():
    rm = ResourceManager()
    c = Component('base')
    r_type = 'column'
    r_names = []
    r_producer = c.producer
    r_dependencies = []

    rg = rm._get_resource_group(r_type, r_names, r_producer, r_dependencies)

    assert rg.type == NULL_RESOURCE_TYPE
    assert rg.names == [f'{NULL_RESOURCE_TYPE}.0']
    assert rg.producer == c.producer
    assert not rg.dependencies
    assert list(rg) == rg.names


def test_resource_manager_add_resources_bad_type():
    rm = ResourceManager()
    c = Component('base')
    r_type = 'unknown'
    r_names = [str(i) for i in range(5)]
    r_producer = c.producer
    r_dependencies = []

    with pytest.raises(ResourceError, match='Unknown resource type'):
        rm.add_resources(r_type, r_names, r_producer, r_dependencies)


def test_resource_manager_add_resources_multiple_producers():
    rm = ResourceManager()
    c1 = Component('1')
    c2 = Component('2')
    r_type = 'column'
    r1_names = [str(i) for i in range(5)]
    r2_names = [str(i) for i in range(5, 10)] + ['1']
    r1_producer = c1.producer
    r2_producer = c2.producer
    r_dependencies = []

    rm.add_resources(r_type, r1_names, r1_producer, r_dependencies)
    with pytest.raises(ResourceError, match='producers for column.1'):
        rm.add_resources(r_type, r2_names, r2_producer, r_dependencies)


def test_resource_manager_add_resources():
    rm = ResourceManager()
    for r_type in RESOURCE_TYPES:
        old_names = []
        for i in range(5):
            c = Component(f'r_type_{i}')
            names = [f'r_type_{i}_{j}' for j in range(5)]
            rm.add_resources(r_type, names, c.producer, old_names)
            old_names = names


def test_resource_manager_sorted_nodes_two_node_cycle():
    rm = ResourceManager()
    c = Component('test')

    rm.add_resources('column', ['1'], c.producer, ['stream.2'])
    rm.add_resources('stream', ['2'], c.producer, ['column.1'])

    with pytest.raises(ResourceError, match='cycle'):
        _ = rm.sorted_nodes


def test_resource_manager_sorted_nodes_three_node_cycle():
    rm = ResourceManager()
    c = Component('test')

    rm.add_resources('column', ['1'], c.producer, ['stream.3'])
    rm.add_resources('stream', ['2'], c.producer, ['column.1'])
    rm.add_resources('stream', ['3'], c.producer, ['stream.2'])

    with pytest.raises(ResourceError, match='cycle'):
        _ = rm.sorted_nodes


def test_resource_manager_sorted_nodes_large_cycle():
    rm = ResourceManager()
    c = Component('test')

    for i in range(10):
        rm.add_resources('column', [f'{i}'], c.producer, [f'column.{i%10}'])

    with pytest.raises(ResourceError, match='cycle'):
        _ = rm.sorted_nodes


def test_resource_manager_sorted_nodes_diamond():
    rm = ResourceManager()
    c = Component('test')

    rm.add_resources('column', ['1'], c.producer, [])
    rm.add_resources('column', ['2'], c.producer, ['column.1'])
    rm.add_resources('column', ['3'], c.producer, ['column.1'])
    rm.add_resources('column', ['4'], c.producer, ['column.2', 'column.3'])

    n = [str(node) for node in rm.sorted_nodes]

    assert n.index('(column.1)') < n.index('(column.2)')
    assert n.index('(column.1)') < n.index('(column.3)')
    assert n.index('(column.1)') < n.index('(column.4)')

    assert n.index('(column.2)') < n.index('(column.4)')
    assert n.index('(column.3)') < n.index('(column.4)')



import json
from pathlib import Path
import random

import numpy as np
import pandas as pd
import pytest
import tables
from tables.nodes import filenode

from vivarium.testing_utilities import build_table
from vivarium.framework.artifact import hdf


_KEYS = ['population.age_bins',
         'population.structure',
         'population.theoretical_minimum_risk_life_expectancy',
         'cause.all_causes.restrictions',
         'metadata.versions',
         'metadata.locations',
         'metadata.keyspace']


@pytest.fixture
def hdf_keys():
    return _KEYS


@pytest.fixture(params=_KEYS)
def hdf_key(request):
    return request.param


@pytest.fixture(params=['totally.new.thing', 'other.new_thing', 'cause.sort_of_new', 'cause.also.new',
                        'cause.all_cause.kind_of_new'])
def mock_key(request):
    return hdf.EntityKey(request.param)


@pytest.fixture(params=[[], {}, ['data'], {'thing': 'value'}, 'bananas'])
def json_data(request):
    return request.param


def test_touch_no_file(mocker):
    path = Path('not/an/existing/path.hdf')
    tables_mock = mocker.patch("vivarium.framework.artifact.hdf.tables")

    hdf.touch(path)
    tables_mock.open_file.assert_called_once_with(str(path), mode='w')
    tables_mock.reset_mock()


def test_touch_exists_but_not_hdf_file_path(hdf_file_path):
    dir_path = Path(hdf_file_path).parent
    with pytest.raises(ValueError):
        hdf.touch(dir_path)
    non_hdf_path = Path(hdf_file_path).parent / 'test.txt'
    with pytest.raises(ValueError):
        hdf.touch(non_hdf_path)


def test_touch_existing_file(tmpdir):
    path = f'{str(tmpdir)}/test.hdf'

    hdf.touch(path)
    hdf.write(path, hdf.EntityKey('test.key'), 'data')
    assert hdf.get_keys(path) == ['test.key']

    # should wipe out and make it again
    hdf.touch(path)
    assert hdf.get_keys(path) == []


def test_write_df(hdf_file_path, mock_key, mocker):
    df_mock = mocker.patch('vivarium.framework.artifact.hdf._write_pandas_data')
    data = pd.DataFrame(np.random.random((10, 3)), columns=['a', 'b', 'c'], index=range(10))

    hdf.write(hdf_file_path, mock_key, data)

    df_mock.assert_called_once_with(hdf_file_path, mock_key, data)


def test_write_json(hdf_file_path, mock_key, json_data, mocker):
    json_mock = mocker.patch('vivarium.framework.artifact.hdf._write_json_blob')
    hdf.write(hdf_file_path, mock_key, json_data)
    json_mock.assert_called_once_with(hdf_file_path, mock_key, json_data)


def test_load(hdf_file_path, hdf_key):
    key = hdf.EntityKey(hdf_key)
    data = hdf.load(hdf_file_path, key, filter_terms=None, column_filters=None)
    if 'restrictions' in key or 'versions' in key:
        assert isinstance(data, dict)
    elif 'metadata' in key:
        assert isinstance(data, list)
    else:
        assert isinstance(data, pd.DataFrame)


def test_load_with_invalid_filters(hdf_file_path, hdf_key):
    key = hdf.EntityKey(hdf_key)
    data = hdf.load(hdf_file_path, key, filter_terms=["fake_filter==0"], column_filters=None)
    if 'restrictions' in key or 'versions' in key:
        assert isinstance(data, dict)
    elif 'metadata' in key:
        assert isinstance(data, list)
    else:
        assert isinstance(data, pd.DataFrame)


def test_load_with_valid_filters(hdf_file_path, hdf_key):
    key = hdf.EntityKey(hdf_key)
    data = hdf.load(hdf_file_path, key, filter_terms=["year == 2006"], column_filters=None)
    if 'restrictions' in key or 'versions' in key:
        assert isinstance(data, dict)
    elif 'metadata' in key:
        assert isinstance(data, list)
    else:
        assert isinstance(data, pd.DataFrame)
        if 'year' in data.columns:
            assert set(data.year) == {2006}


def test_load_filter_empty_data_frame_index(hdf_file_path):
    key = hdf.EntityKey('cause.test.prevalence')
    data = pd.DataFrame(data={'age': range(10),
                              'year': range(10),
                              'draw': range(10)})
    data = data.set_index(list(data.columns))

    hdf._write_pandas_data(hdf_file_path, key, data)
    loaded_data = hdf.load(hdf_file_path, key, filter_terms=['year == 4'], column_filters=None)
    loaded_data = loaded_data.reset_index()
    assert loaded_data.year.unique() == 4


def test_remove(hdf_file_path, hdf_key):
    key = hdf.EntityKey(hdf_key)
    hdf.remove(hdf_file_path, key)
    with tables.open_file(str(hdf_file_path)) as file:
        assert key.path not in file


def test_get_keys(hdf_file_path, hdf_keys):
    assert sorted(hdf.get_keys(hdf_file_path)) == sorted(hdf_keys)


def test_write_json_blob(hdf_file_path, mock_key, json_data):
    hdf._write_json_blob(hdf_file_path, mock_key, json_data)

    with tables.open_file(str(hdf_file_path)) as file:
        node = file.get_node(mock_key.path)
        with filenode.open_node(node) as file_node:
            data = json.load(file_node)
            assert data == json_data


def test_write_empty_data_frame(hdf_file_path):
    key = hdf.EntityKey('cause.test.prevalence')
    data = pd.DataFrame(columns=('age', 'year', 'sex', 'draw', 'location', 'value'))

    with pytest.raises(ValueError):
        hdf._write_pandas_data(hdf_file_path, key, data)


def test_write_empty_data_frame_index(hdf_file_path):
    key = hdf.EntityKey('cause.test.prevalence')
    data = pd.DataFrame(data={'age': range(10),
                              'year': range(10),
                              'draw': range(10)})
    data = data.set_index(list(data.columns))

    hdf._write_pandas_data(hdf_file_path, key, data)
    written_data = pd.read_hdf(hdf_file_path, key.path)
    written_data = written_data.set_index(list(written_data))  # write resets index. only calling load undoes it
    assert written_data.equals(data)


def test_write_load_empty_data_frame_index(hdf_file_path):
    key = hdf.EntityKey('cause.test.prevalence')
    data = pd.DataFrame(data={'age': range(10),
                              'year': range(10),
                              'draw': range(10)})
    data = data.set_index(list(data.columns))

    hdf._write_pandas_data(hdf_file_path, key, data)
    loaded_data = hdf.load(hdf_file_path, key, filter_terms=None, column_filters=None)
    assert loaded_data.equals(data)


def test_write_data_frame(hdf_file_path):
    key = hdf.EntityKey('cause.test.prevalence')
    data = build_table([lambda *args, **kwargs: random.choice([0, 1]), "Kenya", 1],
                       2005, 2010, columns=('age', 'year', 'sex', 'draw', 'location', 'value'))

    non_val_columns = data.columns.difference({'value'})
    data = data.set_index(list(non_val_columns))

    hdf._write_pandas_data(hdf_file_path, key, data)

    written_data = pd.read_hdf(hdf_file_path, key.path)
    assert written_data.equals(data)

    filter_terms = ['draw == 0']
    written_data = pd.read_hdf(hdf_file_path, key.path, where=filter_terms)
    assert written_data.equals(data.xs(0, level='draw', drop_level=False))


def test_get_keys_private(hdf_file, hdf_keys):
    assert sorted(hdf._get_keys(hdf_file.root)) == sorted(hdf_keys)


def test_get_node_name(hdf_file, hdf_key):
    key = hdf.EntityKey(hdf_key)
    assert hdf._get_node_name(hdf_file.get_node(key.path)) == key.measure


def test_get_valid_filter_terms_all_invalid(hdf_key, hdf_file):
    node = hdf_file.get_node(hdf.EntityKey(hdf_key).path)
    if not isinstance(node, tables.earray.EArray):
        columns = node.table.colnames
        invalid_filter_terms = _construct_no_valid_filters(columns)
        assert hdf._get_valid_filter_terms(invalid_filter_terms, columns) is None


def test_get_valid_filter_terms_all_valid(hdf_key, hdf_file):
    node = hdf_file.get_node(hdf.EntityKey(hdf_key).path)
    if not isinstance(node, tables.earray.EArray):
        columns = node.table.colnames
        valid_filter_terms = _construct_all_valid_filters(columns)
        assert set(hdf._get_valid_filter_terms(valid_filter_terms, columns)) == set(valid_filter_terms)


def test_get_valid_filter_terms_some_valid(hdf_key, hdf_file):
    node = hdf_file.get_node(hdf.EntityKey(hdf_key).path)
    if not isinstance(node, tables.earray.EArray):
        columns = node.table.colnames
        invalid_filter_terms = _construct_no_valid_filters(columns)
        valid_filter_terms = _construct_all_valid_filters(columns)
        all_terms = invalid_filter_terms + valid_filter_terms
        result = hdf._get_valid_filter_terms(all_terms, columns)
        assert set(result) == set(valid_filter_terms)


def test_get_valid_filter_terms_no_terms():
    assert hdf._get_valid_filter_terms(None, []) is None


def _construct_no_valid_filters(columns):
    fake_cols = [c[1:] for c in columns] # strip out the first char to make a list of all fake cols
    terms = [c + ' <= 0' for c in fake_cols]
    return _complicate_terms_to_parse(terms)


def _construct_all_valid_filters(columns):
    terms = [c + '=0' for c in columns] # assume c is numeric - we won't actually apply filter
    return _complicate_terms_to_parse(terms)


def _complicate_terms_to_parse(terms):
    n_terms = len(terms)
    if n_terms > 1:
        # throw in some parens and ifs/ands
        term_1 = '(' + ' & '.join(terms[:(n_terms//2+n_terms % 2)]) + ')'
        term_2 = '(' + ' | '.join(terms[(n_terms//2+n_terms % 2):]) + ')'
        terms = [term_1, term_2] + terms
    return ['(' + t + ')' for t in terms]


def test_EntityKey_init_failure():
    bad_keys = ['hello', 'a.b.c.d', '', '.', '.coconut', 'a.', 'a..c']

    for k in bad_keys:
        error_msg = f'Invalid format for HDF key: {k}. Acceptable formats are "type.name.measure" and "type.measure"'
        with pytest.raises(ValueError, match=error_msg):
            hdf.EntityKey(k)


def test_EntityKey_no_name():
    type_ = 'population'
    measure = 'structure'
    key = hdf.EntityKey(f'{type_}.{measure}')

    assert key.type == type_
    assert key.name == ''
    assert key.measure == measure
    assert key.group_prefix == '/'
    assert key.group_name == type_
    assert key.group == f'/{type_}'
    assert key.path == f'/{type_}/{measure}'
    assert key.with_measure('age_groups') == hdf.EntityKey('population.age_groups')


def test_EntityKey_with_name():
    type_ = 'cause'
    name = 'diarrheal_diseases'
    measure = 'incidence'
    key = hdf.EntityKey(f'{type_}.{name}.{measure}')

    assert key.type == type_
    assert key.name == name
    assert key.measure == measure
    assert key.group_prefix == f'/{type_}'
    assert key.group_name == name
    assert key.group == f'/{type_}/{name}'
    assert key.path == f'/{type_}/{name}/{measure}'
    assert key.with_measure('prevalence') == hdf.EntityKey(f'{type_}.{name}.prevalence')


def test_entity_key_equality():

    type_ = 'cause'
    name = 'diarrheal_diseases'
    measure = 'incidence'
    string = f'{type_}.{name}.{measure}'
    key = hdf.EntityKey(string)

    class NonString:
        def __str__(self):
            return string

    nonstring = NonString()

    assert key == string, 'Comparision using __eq__ between string object and equivalent EntityKey failed'
    assert not (key != string), 'Comparision using __ne__ between string object and equivalent EntityKey failed'
    assert key != nonstring, 'Comparision using __eq__ between non-string object and equivalent EntityKey failed'
    assert not (key == nonstring), 'Comparision using __ne__ between non-string object and equivalent EntityKey failed'

    measure = 'prevalence'
    string = f'{type_}.{name}.{measure}'

    assert key != string, 'Comparision using __eq__ between string object and different EntityKey failed'
    assert not (key == string), 'Comparision using __ne__ between string object and different EntityKey failed'
    assert key != nonstring, 'Comparision using __eq__ between non-string object and different EntityKey failed'
    assert not (key == nonstring), 'Comparision using __ne__ between non-string object and different EntityKey failed'

